// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "graphicssystemfactory.h"
#include "mtxlib.h"
#include "keyboard.h"
#include "frustum.h"
#include "filenames.h"
#include "application.h"
#include "get_distance.h"
#include "gettime.h"
#include "audiosystemopenal.h"
#include "hud_renderer.h"
#include "hud.h"
#include "font.h"
#include <string>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <deque>
#include "object_renderer.h"
#include "error.h"
#include "settings.h"

#if defined(USE_SOUND)
ALuint explosion_sample;
ALuint active_sources[100];
#endif // defined(USE_SOUND)

extern int g_width, g_height, g_bpp, g_flags;

TEXTURE_HANDLE font_texture01;

// Constructor
Application::Application()
{
    try
    {
        IGS = GraphicsSystemFactory::CreateGraphicsSystem(GraphicsSystemFactory::GRAPHICS_OPENGL);

        hud_misc = IGS->CreateTexture(data_dir() + slash() + "hud_misc.png", false);
        font_texture01 = IGS->CreateTexture(data_dir() + slash() + "font01.png", false);
/*
        Timer timer;
        timer.Go();
*/
        level = new Level(IGS, this);
        editor = NULL;
/*
        std::cout << "Load time: " << timer.GetElapsedTime() << std::endl;
*/
#if defined(USE_SOUND)
        al = new OpenAL();

        if (al->Init() == 1)
        {
            audio_available = false;
            std::cout << "OpenAL Initialisation failed" << std::endl;
        }
        else
        {
            audio_available = true;
        }

        if (audio_available)
        {
            alDistanceModel(AL_INVERSE_DISTANCE);

            explosion_sample = al->CreateBuffer();

            if (al->LoadBuffer(explosion_sample,
                data_dir() + slash() + "sound" + slash() + "explosion.ogg") == 1)
            {
                std::cout << "Failed to load ogg file" << std::endl;
                exit(1);
            }

            for (int i = 0; i < 100; i++)
            {
                active_sources[i] = 0;
            }
        }
#endif // defined(USE_SOUND)
    }
    catch (std::string error_message)
    {
        Error(error_message);
    }
}

// Destructor
Application::~Application()
{
#if defined(USE_SOUND)
    if (audio_available)
    {
        al->DeleteBuffer(explosion_sample);

        al->Shutdown();
    }
    delete al;
#endif // defined(USE_SOUND)

    IGS->DeleteTexture(hud_misc);
    IGS->DeleteTexture(font_texture01);

    delete level;

    if (editor != NULL)
    {
        delete editor;
    }

    delete IGS;
}

// Main game Update function. Call this once per frame.
void Application::Update()
{

    // FPS counter
    static long framecount = 0;
    framecount++;
    static long start_time;

    long frame_time = GetTime() - start_time;
    start_time = GetTime();

    float frame_rate = 0;
    if (framecount > 1)
    {
        frame_rate = 1000.0f / frame_time;
    }

    static std::deque<float> previous_framerates;

    previous_framerates.push_back(frame_rate);
    if (previous_framerates.size() > 5)
    {
        previous_framerates.pop_front();
    }

    float rolling_average = 0;
    for (unsigned int fr = 0; fr < previous_framerates.size(); fr++)
    {
        rolling_average += previous_framerates[fr];
    }
    rolling_average /= float(previous_framerates.size());

    std::stringstream frame_rate_sstr;

    frame_rate_sstr.setf(std::ios::fixed, std::ios::floatfield);
    frame_rate_sstr << std::setprecision(1) << rolling_average << " FPS";

    try
    {
        static int screenshot_number = 0;
        if (Keyboard::Instance().KeyPressed(SDLK_o))
        {
            screenshot_number++;
            std::stringstream screenshot_name;
            screenshot_name << "slickworm" << screenshot_number << ".jpeg";

            IGS->Screenshot(screenshot_name.str());
        }

        IGS->ShowWireframe(Keyboard::Instance().QueryCurrentValueOfSwitch(SDLK_F5));
        IGS->ShowShadowVolumes(Keyboard::Instance().QueryCurrentValueOfSwitch(SDLK_F6));

        static bool editor_mode = false;
        static bool first_run = true;
        bool editor_mode_new = Keyboard::Instance().QueryCurrentValueOfSwitch(SDLK_F9);

        if (editor_mode == false && editor_mode_new == true)
        {
            // Show the mouse cursor
            SDL_ShowCursor(1);

            if (first_run == true)
            {
                // Enter editor mode
                editor = new Editor(IGS, level);

                first_run = false;
            }

            editor_mode = true;
        }
        else if (editor_mode == true && editor_mode_new == false)
        {
            // Hide the mouse cursor
            SDL_ShowCursor(0);

            editor_mode = false;
        }

        level->Update();



        // Initialise the matrices
        matrix44 world_matrix = IdentityMatrix44();

        IGS->SetTransformationMatrix(OBJECT_TO_WORLD_TRANSFORM, world_matrix);

        vector3 camera, lookat, up;

        if (editor_mode == false)
        {
            camera = level->GetCamera();
            lookat = level->GetLookat();
            up = level->GetUp();
        }
        else // (editor_mode == true)
        {
            camera = editor->GetCamera();
            lookat = editor->GetLookat();
            up = editor->GetUp();
        }

        matrix44 view_matrix = LookAtMatrix44(camera, lookat, up);


        IGS->SetTransformationMatrix(VIEW_TRANSFORM, view_matrix);

#if defined(USE_SOUND)
        if (audio_available)
        {
            al->SetListenerPosition(camera);
            al->SetListenerOrientation(lookat, up);
            al->SetListenerVelocity(vector3(0.0f, 0.0f, 0.0f));

            for (int s = 0; s < 100; s++)
            {
                if (s != 0 && al->SourceIsPlaying(active_sources[s]) == false)
                {
                    al->DeleteSource(active_sources[s]);
                    active_sources[s] = 0;
                }
            }
        }
#endif // defined(USE_SOUND)

        //matrix44 perspective_matrix = PerspectiveMatrix44(45.0f/2.0f, 1.1f*2.0, 0.1f, 2000.0f);
        matrix44 perspective_matrix = PerspectiveMatrix44(45.0f, 1.1f, 0.1f, 2000.0f);

        IGS->SetTransformationMatrix(PROJECTION_TRANSFORM, perspective_matrix);

        // Extract the 6 planes of the frustum
        ExtractFrustum(view_matrix, perspective_matrix);

        ObjectRenderer::Instance().Begin(IGS, camera, view_matrix);



        IGS->Clear(COLOUR(0.0f, 0.0f, 0.0f, 1.0f));

        IGS->Begin();

        IGS->FogParams(0.001f, COLOUR(0.0, 0.0, 0.0, 1.0));
        IGS->EnableFog(false);


        level->Render1(camera);

        Light ambient;
        ambient.type = Light::AMBIENT;
        ambient.colour = COLOUR(0.3f, 0.3f, 0.3f, 1.0f);
        ObjectRenderer::Instance().AddLight(ambient);

        Light directional;

        directional.type = Light::DIRECTIONAL;
        directional.direction = vector3(0.424f, -0.8f, 0.424f).normalize();
        directional.colour = COLOUR(0.5f, 0.5f, 0.5f, 1.0f);
        ObjectRenderer::Instance().AddLight(directional);

/*
        directional.type = Light::DIRECTIONAL;
        directional.direction = vector3(0.124f, -0.18f, -0.324f).normalize();
        directional.colour = COLOUR(0.5f, 0.2f, 0.1f, 1.0f);
        ObjectRenderer::Instance().AddLight(directional);
*/


        if (editor_mode == true)
        {
            editor->RenderObject();
        }

        ObjectRenderer::Instance().RenderScene();

        
        IGS->SetBlendMode(BLEND_DISABLE);
        IGS->SetColour(COLOUR(1.0f, 1.0f, 1.0f));

        level->Render2(camera);


        if (editor_mode == true)
        {
            editor->Render(perspective_matrix, view_matrix);
        }





        ObjectRenderer::Instance().RenderTransparentObjects();


        // Draw the HUD
        DisplayText(ARIAL_BLACK, 5, 580, 0.3f, frame_rate_sstr.str().c_str(),
            font_texture01, IGS, COLOUR(1.0f, 1.0f, 1.0f, 1.0f));


/*
        IGS->SetTexture(0, hud_misc);

        draw_health_bar(IGS, 20, 20, BAR_RED);

        HUD_Renderer::Instance().Draw2D(IGS, 10.0f, 3.0f, 62.0f, 19.0f,
            0.0 / 256.0, 32.0 / 256.0, 52.0 / 256.0, 48.0 / 256.0, COLOUR(1.0f, 1.0f, 0.25f, 1.0f));
*/
/*
        HUD_Renderer::Instance().Draw2D(IGS, 100, 100, 100 + 91, 100 + 59,
            159.0 / 256.0, 31.0 / 256.0, 250.0 / 256.0, 90.0 / 256.0, COLOUR(1.0f, 1.0f, 1.0f, 1.0f));
        DisplayText(ARIAL_BLACK, 105, 102, 0.3f, "Hello",
            font_texture01, IGS, COLOUR(0.0f, 0.0f, 0.0f, 1.0f));
*/
/*
        IGS->CopyScreenToTexture(hud_misc, 0, 0, 512 / 2, 256 / 2);
        IGS->SetTexture(0, hud_misc);
        HUD_Renderer::Instance().Draw2D(IGS, 100, 100, 300, 240,
            0.0 / 256.0, 0.0 / 256.0, 256.0 / 256.0, 256.0 / 256.0, COLOUR(1.0f, 1.0f, 1.0f, 1.0f));
*/


        IGS->SetTexture(0, hud_misc);

        static bool show_get_into_vehicle_message = false;

        if (level->PlayerNearBuggy())
        {
            show_get_into_vehicle_message = true;
        }

        if (show_get_into_vehicle_message)
        {
            static long start_fade_time = GetTime();
            long elapsed_fade_time = GetTime() - start_fade_time;

            if (elapsed_fade_time < 8000)
            {
                IGS->SetTexture(0, hud_misc);
                draw_message("Press [Return] to enter the vehicle", font_texture01, IGS, elapsed_fade_time);
            }
        }

        static bool show_buggy_overturned_message = false;

        if (level->BuggyFlippedOver())
        {
            show_buggy_overturned_message = true;
        }

        if (show_buggy_overturned_message)
        {
            static long start_fade_time = GetTime();
            long elapsed_fade_time = GetTime() - start_fade_time;

            if (elapsed_fade_time < 8000)
            {
                IGS->SetTexture(0, hud_misc);
                draw_message("Press [R] to reset the vehicle", font_texture01, IGS, elapsed_fade_time);
            }
        }


        // Functions dealing with changing the current resolution.
        if (Keyboard::Instance().QueryCurrentValueOfSwitch(SDLK_F12))
        {
            static int current_mode = 0;

            // Get available fullscreen/hardware modes
            SDL_Rect** modes=SDL_ListModes(NULL, g_flags);

            DisplayText(ARIAL_BLACK, 150, 100, 0.3f, "Use the UP and DOWN arrow keys to highlight a mode.",
                        font_texture01, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.7f));
            DisplayText(ARIAL_BLACK, 150, 113, 0.3f, "Press RETURN to select that mode.",
                        font_texture01, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.7f));

            DisplayText(ARIAL_BLACK, 150, 113+26+26, 0.3f, "Press F12 again to remove this text.",
                        font_texture01, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.7f));

            DisplayText(ARIAL_BLACK, 150, 113+26, 0.3f, "Press F11 to remember current mode on startup.",
                        font_texture01, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.7f));

            int mode_count = 0;

            for (int mode = 0; modes[mode]; ++mode)
            {
                std::stringstream mode_sstr;
                mode_sstr << modes[mode]->w << "x" << modes[mode]->h;

                if (mode != current_mode)
                {
                    DisplayText(ARIAL_BLACK, 20, 13 * mode, 0.3f, mode_sstr.str().c_str(),
                        font_texture01, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.7f));
                }
                else
                {
                    DisplayText(ARIAL_BLACK, 20, 13 * mode, 0.3f, mode_sstr.str().c_str(),
                        font_texture01, IGS, COLOUR(1.0f, 0.2f, 0.1f, 1.0f));
                }
                mode_count++;
            }

            static long last_pressed_up;
            static long last_pressed_down;

            if (Keyboard::Instance().KeyPressed(SDLK_UP))
            {
                if (GetTime() - last_pressed_up > 50)
                {
                    current_mode--;
                    if (current_mode <= 0) current_mode = 0;
                    last_pressed_up = GetTime();
                }
            }
            if (Keyboard::Instance().KeyPressed(SDLK_DOWN))
            {
                if (GetTime() - last_pressed_down > 50)
                {
                    current_mode++;
                    if (current_mode >= mode_count) current_mode = mode_count - 1;
                    last_pressed_down = GetTime();
                }
            }

            if (Keyboard::Instance().KeyPressed(SDLK_RETURN))
            {
                g_width = modes[current_mode]->w;
                g_height = modes[current_mode]->h;
                SDL_SetVideoMode(g_width, g_height, g_bpp, g_flags);
                IGS->SetViewportSize(g_width, g_height);
                IGS->RestoreTextures();
            }

            if (Keyboard::Instance().KeyPressed(SDLK_F11))
            {
                Settings::Instance().screen_width = g_width;
                Settings::Instance().screen_height = g_height;

                Settings::Instance().Save();
            }
        }










        IGS->End();

        IGS->Display();

        SleepMS(0);
    }
    catch (std::string error_message)
    {
        Error(error_message);
    }
}

void Application::SoundEffect(int effect_number, vector3 location)
{
    // Sound effect
    #if defined(USE_SOUND)
    if (audio_available)
    {
        ALuint vorbsource = al->CreateSource();

        alSourcei(vorbsource, AL_BUFFER, explosion_sample);
        alSourcei(vorbsource, AL_LOOPING, AL_FALSE);
        alSourcef(vorbsource, AL_ROLLOFF_FACTOR, 0.1f);

        alSourcePlay(vorbsource);

        al->SetSourcePosition(vorbsource, location);

        int s = 0;
        while (active_sources[s] != 0)
        {
            s++;
        };
        active_sources[s] = vorbsource;
    }
    #endif // defined(USE_SOUND)
}





