// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <string>
#include "buggy.h"
#include "keyboard.h"
#include "modelfactory.h"
#include "graphicssystem.h"
#include "drawstuff.h"
#include "object_renderer.h"
#include "filenames.h"
#include <iostream>

Buggy::Buggy(dWorldID world2, dSpaceID space2, IGraphicsSystem* IGS,
    std::string buggy_directory, std::string chassis_filename,
    std::string wheel_filename, vector3 initial_position)
{
    OdeThread::ode_mutex_lock();

    world = world2;
    space = space2;
    IGS   = IGS;

    // Load the 3d models for the wheel and chassis.
    chassis = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
    wheel   = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);

    chassis->Load(buggy_directory, chassis_filename, IGS);
    wheel->Load(buggy_directory, wheel_filename, IGS);
    //wheel->Load(data_dir() + slash() + "objects" + slash(), "crate.txt", IGS);

    dMass m;

    // Create a space to hold the wheels and chassis of the buggy
    // This space will be nested in the main space.
    buggy_space = dSimpleSpaceCreate(space);

    // Create the chassis body
    body_c[0] = dBodyCreate(world);
    dBodySetPosition(body_c[0], initial_position.x, initial_position.y + STARTY, initial_position.z);
    dMassSetBox (&m, 1.0 , LENGTH, HEIGHT, WIDTH);
    dMassAdjust (&m, 7);
    dBodySetMass (body_c[0], &m);
/*
    box = dCreateBox (buggy_space, LENGTH, HEIGHT, WIDTH);
    dGeomSetBody (box, body_c[0]);
    dBodySetData(body_c[0], (void *)BODY_BUGGY);
*/

    int i;

    // wheel bodies
    for (i = 0; i < 4; i++)
    {
        body_w[i] = dBodyCreate (world);
        dQuaternion q;
        dQFromAxisAndAngle (q, 0 , 0 , 1, 0.0);
        dBodySetQuaternion (body_w[i], q);
        dMassSetSphere (&m, 1.0, RADIUS);
        dMassAdjust (&m, 1);
        sphere[i] = dCreateSphere (buggy_space, RADIUS);
        dBodySetMass (body_w[i], &m);
        dGeomSetBody (sphere[i], body_w[i]);
    }

    dBodySetPosition (body_w[0],initial_position.x + 0.5*LENGTH, initial_position.y, initial_position.z - 0.5 * WIDTH);
    dBodySetPosition (body_w[1],initial_position.x + 0.5*LENGTH, initial_position.y, initial_position.z + 0.5 * WIDTH);
    dBodySetPosition (body_w[2],initial_position.x - 0.5*LENGTH, initial_position.y, initial_position.z - 0.5 * WIDTH);
    dBodySetPosition (body_w[3],initial_position.x - 0.5*LENGTH, initial_position.y, initial_position.z + 0.5 * WIDTH);

    // front and back wheel hinges
    for (i = 0; i < 4; i++)
    {
        joint[i] = dJointCreateHinge2 (world, 0);
        dJointAttach (joint[i], body_c[0], body_w[i]);
        const dReal *a = dBodyGetPosition (body_w[i]);
        dJointSetHinge2Anchor(joint[i], a[0], a[1], a[2]);
        dJointSetHinge2Axis1 (joint[i],    0,    1,    0);
        dJointSetHinge2Axis2 (joint[i],    0,    0,    1);

        dJointSetHinge2Param(joint[i], dParamSuspensionERP, 0.25);
        dJointSetHinge2Param(joint[i], dParamSuspensionCFM, 0.004);

        dJointSetHinge2Param(joint[i], dParamVel2, 0);
        dJointSetHinge2Param(joint[i], dParamFMax2, 80);

        // set stops to make sure wheels always stay in alignment
        dJointSetHinge2Param (joint[i],dParamLoStop,0);
        dJointSetHinge2Param (joint[i],dParamHiStop,0);
        dJointSetHinge2Param(joint[i], dParamFMax, 50);
    }

    OdeThread::ode_mutex_unlock();
}

Buggy::~Buggy()
{
    OdeThread::ode_mutex_lock();

    delete chassis;
    delete wheel;

    dJointDestroy(joint[0]);
    dJointDestroy(joint[1]);
    dJointDestroy(joint[2]);
    dJointDestroy(joint[3]);

    dBodyDestroy (body_c[0]);
    dBodyDestroy (body_w[0]);
    dBodyDestroy (body_w[1]);
    dBodyDestroy (body_w[2]);
    dBodyDestroy (body_w[3]);

    dSpaceDestroy(buggy_space);

    OdeThread::ode_mutex_unlock();
}

void Buggy::GetCameraPosition(vector3 &position, vector3 &lookat, vector3 &up, bool third_person)
{
    OdeThread::ode_mutex_lock();

    if (third_person)
    {
        // ---------------------------------------------------------------------
        // Compute the camera position
        //----------------------------------------------------------------------

        vector3 pos = ConvertToWorldSpace(vector3(-18.0, 5.0, 0.0), body_c[0]);

        const dReal *position2 = dBodyGetPosition(body_c[0]);

        if (pos.y < position2[1]) pos.y = position2[1];

        position = pos;

        // ---------------------------------------------------------------------
        // Compute the lookat point
        //----------------------------------------------------------------------

        lookat = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body_c[0]);

        // ---------------------------------------------------------------------
        // Compute the up vector
        //----------------------------------------------------------------------

        up = vector3(0.0, 1.0, 0.0);

    }

    else // first person camera.
    {
        // ---------------------------------------------------------------------
        // Compute the camera position
        //----------------------------------------------------------------------

        position = ConvertToWorldSpace(vector3(1.0, 2.4, -0.9), body_c[0]);

        // ---------------------------------------------------------------------
        // Compute the lookat point
        //----------------------------------------------------------------------

        lookat = ConvertToWorldSpace(vector3(4.0, 2.4, -0.9), body_c[0]);

        // ---------------------------------------------------------------------
        // Compute the up vector
        //----------------------------------------------------------------------

        vector3 centre = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body_c[0]);
        vector3 higher = ConvertToWorldSpace(vector3(0.0, 1.0, 0.0), body_c[0]);

        up = higher - centre;
    }
    
    OdeThread::ode_mutex_unlock();
}

//! Returns the location of the buggy
vector3 Buggy::GetPosition()
{
    OdeThread::ode_mutex_lock();
    
    vector3 position = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body_c[0]);

    OdeThread::ode_mutex_unlock();

    return position;
}

//! Returns a normal vector corresponding to the direction the buggy is facing
vector3 Buggy::GetDirection()
{
    OdeThread::ode_mutex_lock();

    vector3 centre  = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body_c[0]);
    vector3 at      = ConvertToWorldSpace(vector3(1.0, 0.0, 0.0), body_c[0]);
    vector3 direction = at - centre;
    direction.normalize();

    OdeThread::ode_mutex_unlock();

    return direction;
}

vector3 Buggy::GetPointInWorldSpace(vector3 point)
{
    OdeThread::ode_mutex_lock();

    vector3 world_point  = ConvertToWorldSpace(point, body_c[0]);

    OdeThread::ode_mutex_unlock();

    return world_point;
}

// Updates the buggy according to user input.
void Buggy::Update(int flags)
{
    if (!IsUpsideDown())    // If buggy is not upside-down.
    {
        // motor
        if (flags & BUGGY_FORWARD) speed += 5.0;
        if (flags & BUGGY_REVERSE) speed -= 5.0;
    }
    
    OdeThread::ode_mutex_lock();

    speed /= 1.8f;

    int w;
    for (w = 0; w < 4; w++)
    {
        dJointSetHinge2Param (joint[w], dParamVel2,speed);
        dJointSetHinge2Param (joint[w], dParamFMax2, 0.4);
    }


    // steering
    bool steer_pressed = false;

    if (flags & BUGGY_TURN_LEFT)    { steer = -5.0; steer_pressed = true; }
    if (flags & BUGGY_TURN_RIGHT)   { steer =  5.0; steer_pressed = true; }

    for (w = 0; w < 4; w++)
    {
        if (w < 2)
        {
            dJointSetHinge2Param (joint[w],dParamVel, steer);
        }
        else
        {
            dJointSetHinge2Param (joint[w],dParamVel, -steer);
        }

        if (!steer_pressed)
        {
            dJointSetHinge2Param (joint[w],dParamLoStop,-0.00);
            dJointSetHinge2Param (joint[w],dParamHiStop, 0.00);
        }
        else
        {
            dJointSetHinge2Param (joint[w],dParamLoStop,-0.20);
            dJointSetHinge2Param (joint[w],dParamHiStop, 0.20);
        }

        dJointSetHinge2Param (joint[w],dParamFMax, 2.0);
        dJointSetHinge2Param (joint[w],dParamFudgeFactor, 0.1);
    }

    OdeThread::ode_mutex_unlock();
}

void Buggy::Render()
{
    OdeThread::ode_mutex_lock();

    for (int w = 0; w < 4; w++)
    {
        matrix44 wheel_transform = ConvertToMatrix44(dBodyGetPosition(body_w[w]), dBodyGetRotation(body_w[w]));
        ObjectRenderer::Instance().AddModel(wheel, wheel_transform);
    }
    matrix44 chassis_transform = ConvertToMatrix44(dBodyGetPosition(body_c[0]), dBodyGetRotation(body_c[0]));
    ObjectRenderer::Instance().AddModel(chassis, chassis_transform);

    OdeThread::ode_mutex_unlock();
}

// Returns true if the buggy is upside down.
bool Buggy::IsUpsideDown()
{
    OdeThread::ode_mutex_lock();

    vector3 centre = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body_c[0]);
    vector3 higher = ConvertToWorldSpace(vector3(0.0, 1.0, 0.0), body_c[0]);

    OdeThread::ode_mutex_unlock();
    
    return (higher.y < centre.y);
}
