// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef BUGGY_H
#define BUGGY_H

/*!
    \file buggy.h
    \brief Deals with the wheeled vehicle.
*/

#include "ode_thread.h"
#include "mtxlib.h"
#include "graphicssystem.h"

class IModel;

#define LENGTH  5.0 // chassis length
#define WIDTH   4.0 // chassis width
#define HEIGHT  2.0 // chassis height
#define RADIUS  1.0 // wheel radius
#define STARTY  0.0 // starting height of chassis
#define CMASS   5.0 // chassis mass
#define WMASS   0.5 // wheel mass

// Flags to describe the direction of the buggy
enum
{
    // If this flag is set buggy will move forward
    BUGGY_FORWARD   = 1,
    // If this flag is set buggy will move backward
    BUGGY_REVERSE   = 2,
    // If this flag is set buggy will turn left
    BUGGY_TURN_LEFT = 4,
    // If this flag is set buggy will turn right
    BUGGY_TURN_RIGHT= 8
};

class Buggy
{
public:

    //! Constructor. Loads the buggy model and the wheel models and creates geoms and bodies.
    /*!
        Note that IGS must point to a valid Rendering interface and should not be deconstructed
        before this object.

        Loads a buggy chassis model and a wheel model. Both are assumed to be in the
        buggy_directory directory. Creates 4 geoms and bodies for the wheels and 1 body and geom
        for the chassis.
    */
    Buggy(dWorldID world2, dSpaceID space2, IGraphicsSystem* IGS,
        std::string buggy_directory, std::string chassis_filename, std::string wheel_filename,
        vector3 initial_position);
    ~Buggy();

    //! Returns the camera position
    /*!
        Returns the position and orientation of a camera when this vehicle is being driven.
        \param position the position of the camera
        \param lookat the point the camera is pointed towards.
        \param up the up vector of the camera.
        \param third_person false if camera is a driver's eye view.
    */
    void GetCameraPosition(vector3 &position, vector3 &lookat, vector3 &up, bool third_person);

    //! Returns the location of the buggy
    vector3 GetPosition();

    //! Returns a normal vector corresponding to the direction the buggy is facing
    vector3 GetDirection();
    
    //! Given a point in the buggy's coordinate space return the corresponding point in world space
    vector3 GetPointInWorldSpace(vector3 point);

    //! Updates the buggy according to user input.
    void Update(int flags);

    //! Renders the buggy.
    void Render();

    //! Returns true if the buggy is upside down.
    /*!
        Returns true if the buggy is upside down. False otherwise.
        If the buggy is upside-down it may sometimes self right itself, if for example it
        is rolling over.
    */
    bool IsUpsideDown();

private:

    // The ODE variables necessary to create geoms and bodies.
    dWorldID world;
    dSpaceID space;

    // Bodies and geoms for the buggy chassis and wheels.
    dBodyID body_c[1];
    dBodyID body_w[4];

    dGeomID box;
    dGeomID sphere[4];
    dJointID joint[4];  // joint[0] and [1] are the front wheels.
    dSpaceID buggy_space;

    // A pointer to the rendering interface.
    IGraphicsSystem *IGS;

    IModel* chassis;    // the 3d model representing the buggy's chassis.
    IModel* wheel;      // the 3d model representing one of the buggy's 4 wheels.

    float speed;        // The speed of the vehicle. Positive = forward, 0 = Stopped.
    float steer;        // The steering of the vehicle. Positive = Right, 0 = Straight.
};

#endif // BUGGY_H
