// Slickworm
// Copyright (c) 2002-2003, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "editor_world.h"
#include "filenames.h"
#include "vector_math.h"
#include "font.h"
#include "mouse.h"
#include "keyboard.h"
#include "modelfactory.h"
#include "object_renderer.h"
#include <sstream>
#include <iostream>

EditorWorld::EditorWorld(Level* level, IGraphicsSystem* IGS)
{
    this->IGS = IGS;
    this->level = level;

    font_texture = IGS->CreateTexture(data_dir() + slash() + "font01.png", false);

    object_filenames.push_back("underpass001.txt");
    object_filenames.push_back("level001.txt");
    object_filenames.push_back("streetlamp.txt");

    object_filenames.push_back("building001.txt");
    object_filenames.push_back("building002.txt");
    object_filenames.push_back("building003.txt");
    object_filenames.push_back("building004.txt");
    object_filenames.push_back("building006.txt");


    current_object = 0;
    current_rotation = DegToRad(0.0);

    current_model = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
    current_model->Load(data_dir() + slash() + "world" + slash(), object_filenames[0], IGS);
}

EditorWorld::~EditorWorld()
{
    IGS->DeleteTexture(font_texture);

    delete current_model;
}

// Create a new world
void EditorWorld::New()
{




}

// Saves the world to a slot
void EditorWorld::Save(int number)
{
    std::stringstream world_filename;
#if defined(WIN32)
    world_filename << "..\\world" << number << ".txt";
#else
    world_filename << "world" << number << ".txt";
#endif

    World* world = this->level->GetWorld();

    world->Save(world_filename.str());
}

// Loads the world from a slot
void EditorWorld::Load(int number)
{
    std::stringstream world_filename;
#if defined(WIN32)
    world_filename << "..\\world" << number << ".txt";
#else
    world_filename << "world" << number << ".txt";
#endif

    World* world = this->level->GetWorld();

    world->Load(world_filename.str());
}

// Feeds the world editor the ray corresponding to the mouse cursor projected into 3D.
void EditorWorld::Ray(bool active, vector3 ray_origin, vector3 ray_destination, float horizontal_plane)
{
    static long last_pressed1;
    static long last_pressed2;

    if (Keyboard::Instance().KeyPressed(SDLK_1))
    {
        if (GetTime() - last_pressed1 > 100)
        {
            current_object--;
            current_object %= object_filenames.size();

            last_pressed1 = GetTime();

            delete current_model;
            current_model = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
            current_model->Load(data_dir() + slash() + "world" + slash(), object_filenames[current_object], IGS);
        }
    }
    if (Keyboard::Instance().KeyPressed(SDLK_2))
    {
        if (GetTime() - last_pressed2 > 100)
        {
            current_object++;
            current_object %= object_filenames.size();

            last_pressed2 = GetTime();

            delete current_model;
            current_model = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
            current_model->Load(data_dir() + slash() + "world" + slash(), object_filenames[current_object], IGS);
        }
    }

    std::stringstream object_sstr;
    object_sstr << "Current object: " << object_filenames[current_object];
    DisplayText(ARIAL_BLACK, 5, 20, 0.3f, object_sstr.str().c_str(),
        font_texture, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.8f));

    static long last_pressed3;
    static long last_pressed4;

    if (Keyboard::Instance().KeyPressed(SDLK_3))
    {
        if (GetTime() - last_pressed3 > 50)
        {
            current_rotation-= DegToRad(1.0);

            last_pressed3 = GetTime();
        }
    }
    if (Keyboard::Instance().KeyPressed(SDLK_4))
    {
        if (GetTime() - last_pressed4 > 50)
        {
            current_rotation+= DegToRad(1.0);

            last_pressed4 = GetTime();
        }
    }

    if (Keyboard::Instance().KeyPressed(SDLK_5))
    {
        current_rotation = DegToRad(0.0);
    }
    if (Keyboard::Instance().KeyPressed(SDLK_6))
    {
        current_rotation = DegToRad(90.0);
    }
    if (Keyboard::Instance().KeyPressed(SDLK_7))
    {
        current_rotation = DegToRad(180.0);
    }
    if (Keyboard::Instance().KeyPressed(SDLK_8))
    {
        current_rotation = DegToRad(270.0);
    }

    std::stringstream rotation_sstr;
    rotation_sstr << "Current rotation: " << current_rotation;
    DisplayText(ARIAL_BLACK, 5, 35, 0.3f, rotation_sstr.str().c_str(),
        font_texture, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.8f));


    if (active == true)
    {
        float distance = intersectRayPlane(ray_origin, (ray_destination - ray_origin).normalize(), vector3(0.0f, horizontal_plane, 0.0f), vector3(0.0f, 1.0f, 0.0f));

        vector3 picked = ray_origin + (ray_destination - ray_origin).normalize() * distance;

        int position_x = (int)(picked.x + 0.5);
        int position_y = (int)(picked.y + 0.5);
        int position_z = (int)(picked.z + 0.5);

        position_x -= (position_x - 8) % 16 - 8;
        position_z -= (position_z - 8) % 16 - 8;

        std::stringstream test_sstr;
        test_sstr << "Position: " << "X: " << position_x << "  Y: " << position_y << "  Z: " << position_z;
        DisplayText(ARIAL_BLACK, 5, 5, 0.3f, test_sstr.str().c_str(),
            font_texture, IGS, COLOUR(1.0f, 1.0f, 1.0f, 0.8f));

        current_position = vector3(position_x, position_y, position_z);

        if (Mouse::Instance().isLeftButtonDown())
        {
            World* world = this->level->GetWorld();

            world->AddInstance(object_filenames[current_object], current_position, current_rotation);

        }

    }






}

void EditorWorld::RenderObject()
{
    // Render the object at the cursor.
    matrix44 transform = TranslateMatrix44(current_position.x, current_position.y, current_position.z)
        * RotateRadMatrix44('Y', current_rotation);
    ObjectRenderer::Instance().AddModel(current_model, transform);
}

