// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef GRAPHICS_SYSTEM_H
#define GRAPHICS_SYSTEM_H

#include <vector>
#include <string>
#include "mtxlib.h"
#include "surface.h"

/*!
    \file graphicssystem.h
    \brief File contains abstract interface for a generic Graphics System.
*/


#define VTX_XYZ             0x00000001
#define VTX_NORMAL          0x00000002
#define VTX_COLOUR          0x00000004
#define VTX_UV1             0x00000008
#define VTX_UV2             0x00000010
#define VTX_UV3             0x00000020

//! A structure describing a colour.
class COLOUR
{
public:
    //! Default constructor
    COLOUR() {}
    //! Constructor with initializing values for r, g, b and optinally a
    COLOUR(float inR, float inG, float inB, float inA = 1.0f): r(inR), g(inG), b(inB), a(inA) {}
    //! Constructor with initializing COLOUR
    COLOUR(const COLOUR &c): r(c.r), g(c.g), b(c.b), a(c.a) {}

    float r;    //!< Red component,   0.0f Black        1.0f Red
    float g;    //!< Green component, 0.0f Black        1.0f Green
    float b;    //!< Blue component,  0.0f Black        1.0f Blue
    float a;    //!< Alpha component, 0.0f Transparent  1.0f Opaque
};

struct VERTEX_XYZ
{
    vector3 point;
};

struct VERTEX_XYZ_COLOUR_UV1
{
    vector3 point;
    COLOUR colour;
    float u, v;
};

struct VERTEX_XYZ_NORMAL
{
    vector3 point;
    vector3 normal;
};

struct VERTEX_XYZ_NORMAL_UV1
{
    vector3 point;
    vector3 normal;
    float u, v;
};

struct VERTEX_XYZ_NORMAL_UV1_UV2
{
    vector3 point;
    vector3 normal;
    float u1, v1, u2, v2;
};

struct VERTEX_XYZ_NORMAL_COLOUR_UV1_UV2
{
    vector3 point;
    vector3 normal;
    COLOUR colour;
    float u1, v1, u2, v2;
};

struct VERTEX_XYZ_UV1
{
    vector3 point;
    float u, v;
};

struct VERTEX_XYZ_UV1_UV2
{
    vector3 point;
    float u1, v1, u2, v2;
};



/*!
    An enum that is passed to DrawTriangles to indicate the way the points
    are organised
*/
enum TRIANGLES_TYPE
{
    TRIANGLE_LIST,
    TRIANGLE_STRIP,
    TRIANGLE_FAN
};

/*!
    An enum of types that are passed to SetTransformationMatrix to tell it
    which transformation matrix to change.
*/
enum TRANSFORMATION_TYPE
{
    OBJECT_TO_WORLD_TRANSFORM,  //!< Object to world space transformation
    VIEW_TRANSFORM,             //!< Camera transform
    PROJECTION_TRANSFORM        //!< Projection to screen transform
};

//! A material used when lighting a surface.
struct MATERIAL
{
    COLOUR diffuse;
    COLOUR specular;
    COLOUR emissive;
    float shiny;        // Sharpness of specular highlights. 1.0 == shiny. 0.0 == not shiny.
};

//! Type of light
enum LIGHTTYPE
{
    DIRECTIONAL_LIGHT   //!< Like a point light infinitely far away.
};

//! A structure that describes a light
struct LIGHT
{
    LIGHTTYPE   type;       //!< Type of light e.g. POINT_LIGHT
    COLOUR      diffuse;    //!< Diffuse colour of light
    COLOUR      specular;   //!< Specular colour
    vector3     direction;  //!< Direction. Ignored if POINT_LIGHT
};

//! Texture wrapping modes
enum TEXTURE_MODE
{
    WRAP_TEXTURE,   //!< Repeat textures
    CLAMP_TEXTURE,  //!< Clamp textures at edges

    WRAP_TEXTURE_S, //!< Repeat texture S coordinate
    WRAP_TEXTURE_T, //!< Repeat texture T coordinate
    WRAP_TEXTURE_R, //!< Repeat texture R coordinate
    CLAMP_TEXTURE_S,//!< Clamp texture S coordinate
    CLAMP_TEXTURE_T,//!< Clamp texture T coordinate
    CLAMP_TEXTURE_R,//!< Clamp texture R coordinate
};

//! Texture blend modes for multitexturing
enum TEXTURE_BLEND_MODE
{
    NORMAL,         //!< Normal - no multitexturing
    DOT3_BUMP,      //!< Dot3 bumpmapping
    DETAIL_TEXTURE  //!< Detail texture.
};

enum UV_GENERATION_MODE
{
    UV_SPHERE_MAPPING,
    UV_OBJECT_SPACE,
    UV_DEFAULT
};

//! Culling modes
enum cull_mode
{
    CULL_CW,        //!< Cull clockwise triangles
    CULL_CCW,       //!< Cull counter-clockwise triangles
    CULL_NONE       //!< Don't cull any triangles based on winding order.
};

//! Depth buffer modes
enum depth_buffer_mode
{
    READ_ONLY,      //!< Useful when drawing transparent objects
    READ_AND_WRITE, //!< Normal operation
    IGNORE_DEPTH    //!< Useful for drawing the Heads Up Display (HUD).
};

//! Blending modes
enum blend_mode
{
    SRC_ALPHA_ONE,                  //!< For explosions or flares or flames
    SRC_ALPHA_ONE_MINUS_SRC_ALPHA,  //!< For smoke or dirt
    ONE_ONE,                        //!< Additive blending
    ONE_ONE_MINUS_SRC_COLOR,
    BLEND_DISABLE                   //!< For everything else
};

//! Shadow typess
enum shadow_mode
{
    Z_FAIL_SHADOWS, //!< Z-fail stencil shadows
    Z_PASS_SHADOWS  //!< Z-pass stencil shadows
};

enum texture_type
{
    TEXTURE_2D,
    TEXTURE_CUBEMAP,
    TEXTURE_3D
};

//! An int which indirectly references a vertex buffer.
typedef int VERTEX_BUFFER_HANDLE;

//! An int which indirectly references an index buffer.
typedef int INDEX_BUFFER_HANDLE;

//! An int which indirectly references a texture map.
typedef int TEXTURE_HANDLE;

typedef int CUBEMAP_TEXTURE_HANDLE;
typedef int TEXTURE3D_HANDLE;

//! Abstract interface for a Graphics Substem.
class IGraphicsSystem
{
public:
    //! Initialises the graphics system
    IGraphicsSystem() {};

    //! Deinitialises the graphics system
    virtual ~IGraphicsSystem() {};

    //! Clears the screen
    /*!
        Clears the colour buffer, depth buffer and the stencil buffer (if available).
        \param colour The colour to clear the screen to.
    */
    virtual void Clear(COLOUR colour) = 0;


    virtual void SetViewportSize(int width, int height) = 0;

    //! Begins the scene.
    /*!
        Call this once per frame before calling any of the rendering functions.
    */
    virtual void Begin() = 0;

    //! Ends the scene.
    /*!
        Call this once per frame after calling the rendering functions.
    */
    virtual void End() = 0;

    //! Displays the backbuffer onto the screen.
    /*!
        Updates the front buffer with the contents of the backbuffer.
    */
    virtual void Display() = 0;

    //! Creates a vertex buffer
    /*!
        A vertex buffer is a buffer containing vertices that the implementation may choose
        to store in video memory. Vertex buffers are suitable for storing vertices that
        don't change very often. Use the DrawTrianglesUP and DrawIndexedTrianglesUP for
        non static vertices.
        \param vertex_format bitmask of flags describing vertex format e.g. VTX_XYZ | VTX_NORMAL | VTX_UV1
        \param vertex_list pointer to a group of vertices.
        \param number_of_vertices the number of vertices in vertex_list
        \return a handle to a newly allocated vertex buffer.
    */
    virtual VERTEX_BUFFER_HANDLE CreateVertexList(int vertex_format,
        float* vertex_list, int number_of_vertices) = 0;

    //! Deletes a vertex buffer
    /*!
        \param handle the handle of the vertex buffer to delete.
    */
    virtual void DeleteVertexList(VERTEX_BUFFER_HANDLE handle) = 0;

    //! Creates an index buffer.
    /*!
        Creates an index buffer. An index buffer is an array of indices used when rendering vertices.
        \param index_list a vector of indices to be stored in a newly created buffer
        \return a handle to a newly allocated index buffer.
    */
    virtual INDEX_BUFFER_HANDLE CreateIndexBuffer(std::vector<unsigned short> index_list) = 0;

    //! Deletes an index buffer.
    /*!
        \param handle the handle of the index buffer to delete.
    */
    virtual void DeleteIndexBuffer(INDEX_BUFFER_HANDLE handle) = 0;

    //! Creates a texture map from a PNG file
    /*!
        Creates a texture from a PNG file. PNG file may optionally contain an alpha channel. PNG
        file should be a multiple of 2 in width and height.
        \param texture_filename the name of the file (possibly also including a path)
        \param generate_mipmaps boolean value that determines whether or not to generate mipmaps for the
        texture.
        \param alpha set this to true if the PNG file contains an alpha channel.
        \return a handle to a newly created texture
    */
    virtual TEXTURE_HANDLE CreateTexture(std::string texture_filename, bool generate_mipmaps) = 0;

    //! Creates a texture map from a Surface.
    /*!
        surface should be a Surface with a type of R8G8B8 or R8G8B8A8.
        \param surface the Surface to copy to the new texture.
        \param generate_mipmaps boolean value that determines whether or not to generate mipmaps for the
        texture.
    */
    virtual TEXTURE_HANDLE CreateTexture(Surface surface, bool generate_mipmaps) = 0;

    //! Deletes a texture
    /*!
        Deletes a texture when given its handle
        \param handle the texture to delete.
    */
    virtual void DeleteTexture(TEXTURE_HANDLE handle) = 0;

    virtual CUBEMAP_TEXTURE_HANDLE CreateCubemapTexture(Surface px, Surface nx, Surface py, Surface ny,
        Surface pz, Surface nz) = 0;

    virtual TEXTURE3D_HANDLE CreateTexture3D(unsigned char* buffer, int texture_size) = 0;

    //! Restores all textures
    virtual void RestoreTextures() = 0;


    //! Sets the texture for multitexturing stage.
    /*!
        Sets the active texture for a specified multitexturing stage
        \param stage the stage to set the texture
        \param handle handle of the texture to set
    */
    virtual void SetTexture(int stage, TEXTURE_HANDLE handle, texture_type type = TEXTURE_2D) = 0;

    //! Sets the texture blend mode.
    /*!
        Sets the blend mode to apply to the texture stages. Possible values are NORMAL, DOT3_BUMP
        and DETAIL_TEXTURE
        \param mode the mode to apply to the texture stages
    */
    virtual void SetMultiTextureMode(TEXTURE_BLEND_MODE mode) = 0;

    //! Sets the texture wrapping mode
    /*!
        Sets the texture wrapping mode. Possible modes are WRAP_TEXTURE and CLAMP_TEXTURE
        \param stage the stage to which the wrapping mode applies.
        \param mode the mode to set the specified texture stage to.
    */
    virtual void SetTextureWrappingMode(int stage, TEXTURE_MODE mode) = 0;

    //! Sets the culling mode
    /*!
        Sets which triangle winding order to not draw.
        CULL_CW = Cull clockwise triangles
        CULL_CCW = Cull counter-clockwise triangles
        CULL_NONE = Don't cull triangles based on winding order.
        \param cull the cull mode to use
    */
    virtual void SetCullMode(cull_mode cull) = 0;

    //! Sets which material is used for lighting.
    /*!
        Sets the current material used for lighting.
    */
    virtual void SetMaterial(MATERIAL material) = 0;

    //! Draws triangles from a vertex buffer.
    /*!
        Draws a bunch of triangles in a vertex buffer.
        \param type indicates the way the vertices are organised. Currently only TRIANGLE_LISTs are supported.
        \param vhandle handle of vertex buffer to use
        \param vertex_count number of vertices to render
    */
    virtual void DrawTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count) = 0;

    //! Draws triangles from a vertex buffer and index buffer.
    /*!
        Draws a bunch of triangles in a vertex buffer. An index buffer specifies the vertex rendering order.
        \param type indicates the way the vertices are organised. Currently only TRIANGLE_LISTs are supported.
        \param vhandle handle of vertex buffer to use
        \param vertex_count number of vertices to render
        \param ihandle handle of index buffer to use
        \param index_count number of indices to render
    */
    virtual void DrawIndexedTriangles(TRIANGLES_TYPE type, VERTEX_BUFFER_HANDLE vhandle, int vertex_count,
        INDEX_BUFFER_HANDLE ihandle, int index_count) = 0;

    //! Draws triangles from a vertex array
    /*!
        Draws triangles from an array of vertices.
        \param type indicates the way the vertices are organised. Currently only TRIANGLE_LISTs are supported.
        \param vertex_format bitmask of flags describing vertex format e.g. VTX_XYZ | VTX_NORMAL | VTX_UV1
        \param vertex_array pointer to the array of vertices
        \param vertex_count number of vertices to render
    */
    virtual void DrawTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
        int vertex_count) = 0;

    //! Draws triangles from a vertex array and index array
    /*!
        Draws triangles from an array of vertices and an index array.
        \param type indicates the way the vertices are organised. Currently only TRIANGLE_LISTs are supported.
        \param vertex_format bitmask of flags describing vertex format e.g. VTX_XYZ | VTX_NORMAL | VTX_UV1
        \param vertex_array pointer to the array of vertices
        \param vertex_count number of vertices in array pointed to by vertex_array
        \param index_array pointer to array of indices
        \param index_count number of indices to render
    */
    virtual void DrawIndexedTrianglesUP(TRIANGLES_TYPE type, int vertex_format, float *vertex_array,
        int vertex_count, unsigned short *index_array, int index_count) = 0;

    //! Sets a light
    virtual void SetLight(int light_number, LIGHT light) = 0;

    //! Enables/Disables lighting.
    virtual void EnableLighting(bool enable) = 0;

    //! Enables or disables a light
    virtual void EnableLight(int light_number, bool enable) = 0;

    virtual void SetColour(COLOUR colour) = 0;

    //! Sets a transformation matrix.
    /*!
        Sets a transformation matrix for some part of the rendering pipeline.
        \param type The type of transformation to specify.
        OBJECT_TO_WORLD_TRANSFORM is set for each object in the scene in turn.
        VIEW_TRANSFORM specifies the camera matrix
        PROJECTION_TRANSFORM specifies a matrix that is used to project vertices to the screen
        \param m Matrix used for the specified transformation type.
    */
    virtual void SetTransformationMatrix(TRANSFORMATION_TYPE type, const matrix44& m) = 0;

    //! Returns a transformation matrix.
    /*!
        Returns a transformation matrix used for some part of the rendering pipeline.
        \param type the type of transformation to specify.
        OBJECT_TO_WORLD_TRANSFORM is set for each object in the scene in turn.
        VIEW_TRANSFORM specifies the camera matrix
        PROJECTION_TRANSFORM specifies a matrix that is used to project vertices to the screen
        \return The matrix for the requested transformation.
    */
    virtual matrix44 GetTransformationMatrix(TRANSFORMATION_TYPE type) = 0;

    //! Sets the fog parameters
    /*!
        Sets the fog colour and density. This sets simple distance based vertex fog for the scene.
        \param density Fog density.
        \param colour Fog colour.
    */
    virtual void FogParams(float density, COLOUR colour) = 0;

    //! Enables or disables fog.
    /*!
        \param enable Set to true to enable fog or false to disable fog
    */
    virtual void EnableFog(bool enable) = 0;

    //! Sets the blend mode
    /*!
        \param blend Set this to SRC_ALPHA_ONE, SRC_ALPHA_ONE_MINUS_SRC_ALPHA, ONE_ONE, ONE_ONE_MINUS_SRC_COLOR,
        or BLEND_DISABLE.
    */
    virtual void SetBlendMode(blend_mode blend) = 0;

    //! Sets the depth buffer mode
    /*!
        Sets the depth buffer mode to one of the following
        READ_ONLY       Useful when drawing transparent objects
        READ_AND_WRITE  Normal operation
        IGNORE_DEPTH    Useful for drawing the Heads Up Display (HUD).
        \param depth Set this to READ_ONLY, READ_AND_WRITE or IGNORE_DEPTH
    */
    virtual void SetDepthBufferMode(depth_buffer_mode depth) = 0;

    //! Enables or disables wireframe mode
    /*!
        \param enable set to true to enable wireframe mode or false to disable wireframe mode
    */
    virtual void ShowWireframe(bool enable) = 0;

    //! Enables or disables rendering shadow volumes to colour buffer mode
    /*!
        Enables or disables the showing of shadow volumes. Red == ZFAIL Yellow == ZPASS
        \param enable Set to true to enable shadow volumes mode or false to disable shadow volumes mode
    */
    virtual void ShowShadowVolumes(bool enable) = 0;

    //! Grabs a screenshot and saves it to file.
    /*!
        Captures a screenshot.
        \param filename The name of the file.
        \param plus_alpha if this is true save alpha channel too. filename must be .png if this is true.
    */
    virtual void Screenshot(std::string filename, bool plus_alpha = false) = 0;

    //! Clears the stencil buffer to 0
    /*!
        Clears the stencil buffer. Note that IGraphicsSystem::Clear should also clear the stencil buffer. This
        function should be used if the stencil buffer needs to be cleared more than once per frame.
    */
    virtual void ClearStencil() = 0;

    //! Sets up the stencil buffer for drawing shadow volumes.
    /*!
        Each call to this function must be accompanied by
        a corresponding call to IGraphicsSystem::StencilTestEnd later on.
    */
    virtual void PrepareStencil() = 0;

    //! Sets the cull mode and the stencil operation for the first shadow pass.
    /*!
        \param mode Z_FAIL_SHADOWS or Z_PASS_SHADOWS
    */
    virtual void StencilSetFirstShadowPass(shadow_mode mode) = 0;

    //! Sets the cull mode and the stencil operation for the second shadow pass.
    /*!
        \param mode Z_FAIL_SHADOWS or Z_PASS_SHADOWS
    */
    virtual void StencilSetSecondShadowPass(shadow_mode mode) = 0;

    //! Sets the stencil test so that only areas where the stencil buffer is 0 get rendered to.
    virtual void StencilTestBegin() = 0;

    //! Call this to stop the stencil test
    /*!
        Each call to this function must be accompanied by a corresponding call to IGraphicsSystem::PrepareStencil
        earlier on.
    */
    virtual void StencilTestEnd() = 0;

    virtual void CopyScreenToTexture(TEXTURE_HANDLE handle, int x, int y, int width, int height) = 0;

    // Draws a sprite. The texture to use must be set already before calling this function.
    virtual void Draw2D(VERTEX_XYZ_UV1 top_left, VERTEX_XYZ_UV1 top_right, VERTEX_XYZ_UV1 bottom_left,
        VERTEX_XYZ_UV1 bottom_right, COLOUR c) = 0;

    //! Sets the mode for automatic texture generation
    virtual void SetUVGenerationMode(int stage, UV_GENERATION_MODE mode) = 0;
};

#endif // GRAPHICS_SYSTEM_H
