// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include <fstream>
#include <sstream>
#include "milkshapeascii.h"
#include "mtxlib.h"
#include "filenames.h"
#include "get_distance.h"
#include "transform.h"
#include "frustum.h"
#include "vector_math.h"
#include <iostream>
#include "object_renderer.h"

MilkshapeAscii::MilkshapeAscii()
{
    IGS = NULL;
}

MilkshapeAscii::~MilkshapeAscii()
{
    // Release all the meshes
    std::vector<Mesh>::iterator mesh_itor;
    for (mesh_itor = meshes.begin(); mesh_itor < meshes.end(); mesh_itor++)
    {
        IGS->DeleteVertexList(mesh_itor->vertex_buffer);
        IGS->DeleteIndexBuffer(mesh_itor->index_buffer);
        delete[] mesh_itor->index_array;
        delete[] mesh_itor->vertex_array;
        delete[] mesh_itor->vertex_array_rendered;
    }

    // Release all the textures
    std::vector<MilkShapeMaterial>::iterator m_itor;
    for (m_itor = materials.begin(); m_itor < materials.end(); m_itor++)
    {
        IGS->DeleteTexture(m_itor->texture_handle);
    }
}

int FloatsInFormat(int vertex_format)
{
    int floats_in_format = 0;

    if (vertex_format & VTX_XYZ) floats_in_format += 3;
    if (vertex_format & VTX_NORMAL) floats_in_format += 3;
    if (vertex_format & VTX_COLOUR) floats_in_format += 4;
    if (vertex_format & VTX_UV1) floats_in_format += 2;
    if (vertex_format & VTX_UV2) floats_in_format += 2;
    if (vertex_format & VTX_UV3) floats_in_format += 2;

    return floats_in_format;
}

int VertexAlreadyInList(float* vertex_list, int number_of_vertices, float* vertex, int vertex_format)
{
    int floats_in_format = FloatsInFormat(vertex_format);

    for (int v = 0; v < number_of_vertices; v++)
    {
        int floats_same = 0;

        for (int f = 0; f < floats_in_format; f++)
        {
            if (vertex_list[v * floats_in_format + f] == vertex[f])
                floats_same++;
        }
        if (floats_same == floats_in_format) return v;
    }

    return -1;
}

std::string ReadInStringBetweenQuotes(std::ifstream& input_file)
{
    std::stringstream input;

    char c;
    input_file >> c;
    do
    {
        if (c!='"')
        {
            input << c;
        }
        input_file >> c;

    }while (c != '"' && c != '\n');

    return input.str();
}

//! Loads a model from file.
void MilkshapeAscii::Load(std::string directory, std::string filename, IGraphicsSystem *IGS)
{
    // Store the renderer pointer.
    this->IGS = IGS;

    std::stringstream file;
    file << directory << filename;

    std::ifstream input_file;
    input_file.open(file.str().c_str());
    if (input_file.fail())
    {
        std::stringstream error;
        error << "Failed to open " << file.str();
        throw std::string(error.str());
    }

    std::string s;

    input_file >> s >> s >> s >> s; // Ignore "// Milkshape 3D ASCII"
    input_file >> s;                // Ignore "Frames:"

    input_file >> num_frames;

    input_file >> s;                // Ignore "Frame:"

    input_file >> frame;

    input_file >> s;                // Ignore "Meshes:"

    int number_of_meshes;

    input_file >> number_of_meshes;

    meshes.resize(number_of_meshes);

    //-------------------------------------------------------------------------
    // Read in the Mesh data
    //-------------------------------------------------------------------------
    int m;  // for loop iterator

    for (m = 0; m < number_of_meshes; m++)
    {
        // Get the name of the mesh including quotes
        std::string mesh_name = ReadInStringBetweenQuotes(input_file);

        // Get the flags for the mesh
        input_file >> meshes[m].mflags;

        // Get the material index
        input_file >> meshes[m].material_index;

        // --------------------------------------------------------------------
        // Read the vertices
        // --------------------------------------------------------------------

        // Read in the number of points
        int number_of_points;
        input_file >> number_of_points;

        meshes[m].points.resize(number_of_points);

        for (int p = 0; p < number_of_points; p++)
        {
            // Read in the vertex flags
            input_file >> meshes[m].points[p].vflags;

            // Read in the vertex position
            input_file >> meshes[m].points[p].position.x;
            input_file >> meshes[m].points[p].position.y;
            input_file >> meshes[m].points[p].position.z;

            // .. and the texture coordinates
            input_file >> meshes[m].points[p].u;
            input_file >> meshes[m].points[p].v;

            // ... and the bone index
            input_file >> meshes[m].points[p].bone_index;
        }

        // --------------------------------------------------------------------
        // Read the vertex normals
        // --------------------------------------------------------------------

        // Read in the number of normals
        int number_of_normals;
        input_file >> number_of_normals;

        meshes[m].normals.resize(number_of_normals);

        for (int n = 0; n < number_of_normals; n++)
        {
            // Read in the normal
            input_file >> meshes[m].normals[n].x;
            input_file >> meshes[m].normals[n].y;
            input_file >> meshes[m].normals[n].z;
        }

        // --------------------------------------------------------------------
        // Read the faces
        // --------------------------------------------------------------------

        // Read in the number of faces
        int number_of_faces;
        input_file >> number_of_faces;

        meshes[m].faces.resize(number_of_faces);

        for (int f = 0; f < number_of_faces; f++)
        {
            // Read in the face flag
            input_file >> meshes[m].faces[f].fflags;

            // ... and the vertex indices
            input_file >> meshes[m].faces[f].v1;
            input_file >> meshes[m].faces[f].v2;
            input_file >> meshes[m].faces[f].v3;

            // ... and the normal indices
            input_file >> meshes[m].faces[f].n1;
            input_file >> meshes[m].faces[f].n2;
            input_file >> meshes[m].faces[f].n3;

            // .. and the smoothing group index
            input_file >> meshes[m].faces[f].smoothing_group;
        }
    }

    //-------------------------------------------------------------------------
    // Read in the Materials data
    //-------------------------------------------------------------------------

    input_file >> s;                // Ignore "Materials:" header

    int number_of_materials;

    input_file >> number_of_materials;

    materials.resize(number_of_materials);

    for (m = 0; m < number_of_materials; m++)
    {
        // Read in the material name
        std::string material_name = ReadInStringBetweenQuotes(input_file);

        // Read in the ambient colour data
        float a, r, g, b;

        input_file >> r >> g >> b >> a;

        input_file >> r >> g >> b >> a;

        materials[m].material.diffuse.r = r;
        materials[m].material.diffuse.g = g;
        materials[m].material.diffuse.b = b;
        materials[m].material.diffuse.a = a;

        input_file >> r >> g >> b >> a;

        materials[m].material.specular.r = r;
        materials[m].material.specular.g = g;
        materials[m].material.specular.b = b;
        materials[m].material.specular.a = a;

        input_file >> r >> g >> b >> a;

        materials[m].material.emissive.r = r;
        materials[m].material.emissive.g = g;
        materials[m].material.emissive.b = b;
        materials[m].material.emissive.a = a;

        float shininess, transparency;

        input_file >> shininess >> transparency;

        materials[m].material.shiny = (shininess / 128.0f);

        materials[m].transparency = transparency;

        // Read in the texture map name
        materials[m].texture_name = ReadInStringBetweenQuotes(input_file);

        // Read in the alpha map name but ignore it
        materials[m].alpha_texture_name = ReadInStringBetweenQuotes(input_file);
    }

    //-------------------------------------------------------------------------
    // Read in the Bones data
    //-------------------------------------------------------------------------

    input_file >> s;                // Ignore "Bones:" header

    int number_of_bones;

    input_file >> number_of_bones;

    skeleton.resize(number_of_bones);

    int b;
    for (b = 0; b < number_of_bones; b++)
    {
        // Read in the bone name
        skeleton[b].name = ReadInStringBetweenQuotes(input_file);

        // Read in the parent bone name
        skeleton[b].parent_name = ReadInStringBetweenQuotes(input_file);

        // Read in the bone's position and rotation
        input_file >> skeleton[b].flags >> skeleton[b].position[0] >> skeleton[b].position[1] >> skeleton[b].position[2]
                                        >> skeleton[b].rotation[0] >> skeleton[b].rotation[1] >> skeleton[b].rotation[2];

        // Read in the position key count
        int position_key_count;
        input_file >> position_key_count;

        // Read in the position keys
        skeleton[b].position_keyframes.resize(position_key_count);

        for (int p = 0; p < position_key_count; p++)
        {
            input_file >> skeleton[b].position_keyframes[p].time
                >> skeleton[b].position_keyframes[p].parameter[0]
                >> skeleton[b].position_keyframes[p].parameter[1]
                >> skeleton[b].position_keyframes[p].parameter[2];
        }

        // Read in the rotation key count
        int rotation_key_count;
        input_file >> rotation_key_count;

        // Read in the rotation keys
        skeleton[b].rotation_keyframes.resize(rotation_key_count);

        for (int r = 0; r < rotation_key_count; r++)
        {
            input_file >> skeleton[b].rotation_keyframes[r].time
                >> skeleton[b].rotation_keyframes[r].parameter[0]
                >> skeleton[b].rotation_keyframes[r].parameter[1]
                >> skeleton[b].rotation_keyframes[r].parameter[2];
        }
    }

    input_file.close();

    // For each material
    for (m = 0; m < number_of_materials; m++)
    {
        std::string texture_name = materials[m].texture_name;

        if (texture_name != "")
        {
            // Load the texture
            materials[m].texture_handle = IGS->CreateTexture(directory + remove_dot_slash(texture_name), true);

            // Try loading a dot3 bump map
            materials[m].dot3bump_handle = 0;

//             std::string filename = directory + "dot3_" + texture_name;
//             if (file_exists(filename))
//             {
//                 std::cout << "dot3_" + texture_name << " exists" << std::endl;
//                 materials[m].dot3bump_handle = IGS->CreateTexture(filename, true);
//             }
        }
    }

    // For each mesh
    for (m = 0; m < number_of_meshes; m++)
    {
        materials[meshes[m].material_index].wrap_u = false;
        materials[meshes[m].material_index].wrap_v = false;

        for (unsigned int p = 0; p < meshes[m].points.size(); p++)
        {
            if (meshes[m].points[p].u > 1.0)
            {
                materials[meshes[m].material_index].wrap_u = true;
            }
            if (meshes[m].points[p].v > 1.0)
            {
                materials[meshes[m].material_index].wrap_v = true;
            }
        }
    }

    // For each mesh
    for (m = 0; m < number_of_meshes; m++)
    {
        VERTEX_XYZ_NORMAL_UV1 vertex_list[10000];
        std::vector<unsigned short> indices;
        meshes[m].vertex_count = 0;

        // For each face
        for (unsigned int f = 0; f < meshes[m].faces.size(); f++)
        {
            VERTEX_XYZ_NORMAL_UV1 vertices[3];

            vertices[0].point = meshes[m].points[meshes[m].faces[f].v1].position;
            vertices[0].normal = meshes[m].normals[meshes[m].faces[f].n1];
            vertices[0].u = meshes[m].points[meshes[m].faces[f].v1].u;
            vertices[0].v = meshes[m].points[meshes[m].faces[f].v1].v;

            vertices[1].point = meshes[m].points[meshes[m].faces[f].v2].position;
            vertices[1].normal = meshes[m].normals[meshes[m].faces[f].n2];
            vertices[1].u = meshes[m].points[meshes[m].faces[f].v2].u;
            vertices[1].v = meshes[m].points[meshes[m].faces[f].v2].v;

            vertices[2].point = meshes[m].points[meshes[m].faces[f].v3].position;
            vertices[2].normal = meshes[m].normals[meshes[m].faces[f].n3];
            vertices[2].u = meshes[m].points[meshes[m].faces[f].v3].u;
            vertices[2].v = meshes[m].points[meshes[m].faces[f].v3].v;

            for (int v = 0; v < 3; v++)
            {
                int index = VertexAlreadyInList((float*)vertex_list, meshes[m].vertex_count,
                    (float*)&vertices[v], VTX_XYZ | VTX_NORMAL | VTX_UV1);

                if (index==-1)
                {
                    vertex_list[meshes[m].vertex_count].point = vertices[v].point;
                    vertex_list[meshes[m].vertex_count].normal= vertices[v].normal;
                    vertex_list[meshes[m].vertex_count].u = vertices[v].u;
                    vertex_list[meshes[m].vertex_count].v = vertices[v].v;

                    if (v==0)
                    {
                        meshes[m].point_indices.push_back(meshes[m].faces[f].v1);
                        meshes[m].normal_indices.push_back(meshes[m].faces[f].n1);
                    }
                    if (v==1)
                    {
                        meshes[m].point_indices.push_back(meshes[m].faces[f].v2);
                        meshes[m].normal_indices.push_back(meshes[m].faces[f].n2);
                    }
                    if (v==2)
                    {
                        meshes[m].point_indices.push_back(meshes[m].faces[f].v3);
                        meshes[m].normal_indices.push_back(meshes[m].faces[f].n3);
                    }

                    indices.push_back(meshes[m].vertex_count);
                    meshes[m].vertex_count++;
                }
                else
                {
                    indices.push_back(index);
                }
            }
        }

        if (number_of_materials == 0 || materials[meshes[m].material_index].dot3bump_handle == 0)
        {
            // No bumpmapping
            meshes[m].vertex_format = VTX_XYZ | VTX_NORMAL | VTX_UV1;

            meshes[m].vertex_buffer = IGS->CreateVertexList(meshes[m].vertex_format,
                (float*)vertex_list, meshes[m].vertex_count);

            int buffer_size = meshes[m].vertex_count * sizeof(VERTEX_XYZ_NORMAL_UV1);

            meshes[m].vertex_array = new float[buffer_size / sizeof(float)];
            meshes[m].vertex_array_rendered = new float[buffer_size / sizeof(float)];

            memcpy(meshes[m].vertex_array, vertex_list, buffer_size);
            memcpy(meshes[m].vertex_array_rendered, vertex_list, buffer_size);
        }
        else if (number_of_materials != 0 && materials[meshes[m].material_index].dot3bump_handle != 0)
        {
            VERTEX_XYZ_NORMAL_COLOUR_UV1_UV2 vertex_list2[10000];

            // Bumpmapping. Use extra tex coord pair + primary vertex colour.
            meshes[m].vertex_format = VTX_XYZ | VTX_NORMAL | VTX_COLOUR | VTX_UV1 | VTX_UV2;

            for (int v = 0; v < meshes[m].vertex_count; v++)
            {
                vertex_list2[v].point = vertex_list[v].point;
                vertex_list2[v].normal = vertex_list[v].normal;
                vertex_list2[v].u1 = vertex_list[v].u;
                vertex_list2[v].v1 = vertex_list[v].v;
                vertex_list2[v].u2 = vertex_list[v].u;
                vertex_list2[v].v2 = vertex_list[v].v;
            }

            meshes[m].vertex_buffer = IGS->CreateVertexList(meshes[m].vertex_format,
                (float*)vertex_list2, meshes[m].vertex_count);

            int buffer_size = meshes[m].vertex_count * sizeof(VERTEX_XYZ_NORMAL_COLOUR_UV1_UV2);

            meshes[m].vertex_array = new float[buffer_size / sizeof(float)];
            meshes[m].vertex_array_rendered = new float[buffer_size / sizeof(float)];

            memcpy(meshes[m].vertex_array, vertex_list2, buffer_size);
            memcpy(meshes[m].vertex_array_rendered, vertex_list2, buffer_size);
        }



        meshes[m].index_buffer = IGS->CreateIndexBuffer(indices);

        meshes[m].index_count = indices.size();
        meshes[m].index_array = new unsigned short[meshes[m].index_count];

        for (int i = 0; i < meshes[m].index_count; i++)
        {
            meshes[m].index_array[i] = indices[i];
        }
    }

    // For each bone
    for (b = 0; b < number_of_bones; b++)
    {
        if (skeleton[b].parent_name == "")
        {
            skeleton[b].parent = -1;
        }
        else
        {
            // Loop through the rest of the bones looking for the parent of this one
            for (int b2 = 0; b2 < number_of_bones; b2++)
            {
                if (skeleton[b].parent_name == skeleton[b2].name)
                {
                    skeleton[b].parent = b2;
                }
            }
        }
    }

    // Calls the function Shadowable::build_connectivity_list
    this->build_connectivity_list();

    // Setup Bones
    for (b = 0; b < number_of_bones; b++)
    {
        vector3 vRot;

        vRot[0] = RadToDeg(skeleton[b].rotation[0]);
        vRot[1] = RadToDeg(skeleton[b].rotation[1]);
        vRot[2] = RadToDeg(skeleton[b].rotation[2]);
        skeleton[b].relative = angle_matrix(vRot);
        skeleton[b].relative[3][0] = skeleton[b].position[0];
        skeleton[b].relative[3][1] = skeleton[b].position[1];
        skeleton[b].relative[3][2] = skeleton[b].position[2];
        int parent_bone = skeleton[b].parent;
        if (skeleton[b].parent != -1)
        {
            skeleton[b].absolute = concat_transforms(skeleton[parent_bone].absolute, skeleton[b].relative);
            skeleton[b].final = skeleton[b].absolute;
        }
        else
        {
            skeleton[b].absolute = skeleton[b].relative;
            skeleton[b].final = skeleton[b].relative;
        }
    }

    for (m = 0; m < number_of_meshes; m++)
    {
        int f = FloatsInFormat(meshes[m].vertex_format);

        for (int v = 0; v < meshes[m].vertex_count; v++)
        {
            int bone_index = meshes[m].points[meshes[m].point_indices[v]].bone_index;

            if (bone_index != -1)
            {
                vector3 vec;

                // setup initial vertices
                vec[0] = meshes[m].vertex_array[v * f + 0] - skeleton[bone_index].absolute[3][0];
                vec[1] = meshes[m].vertex_array[v * f + 1] - skeleton[bone_index].absolute[3][1];
                vec[2] = meshes[m].vertex_array[v * f + 2] - skeleton[bone_index].absolute[3][2];
                vec = reverse_rotate(vec, skeleton[bone_index].absolute);

                meshes[m].vertex_array[v * f + 0] = vec[0];
                meshes[m].vertex_array[v * f + 1] = vec[1];
                meshes[m].vertex_array[v * f + 2] = vec[2];

                // setup initial normals
                vec[0] = meshes[m].vertex_array[v * f + 3];
                vec[1] = meshes[m].vertex_array[v * f + 4];
                vec[2] = meshes[m].vertex_array[v * f + 5];
                vec = reverse_rotate(vec, skeleton[bone_index].absolute);

                meshes[m].vertex_array[v * f + 3] = vec[0];
                meshes[m].vertex_array[v * f + 4] = vec[1];
                meshes[m].vertex_array[v * f + 5] = vec[2];
            }
        }
    }

    aabb[0] =  66600000.0f;
    aabb[1] = -66600000.0f;
    aabb[2] =  66600000.0f;
    aabb[3] = -66600000.0f;
    aabb[4] =  66600000.0f;
    aabb[5] = -66600000.0f;

    // For each mesh
    for (m = 0; m < number_of_meshes; m++)
    {
        // For each point
        for (unsigned int p = 0; p < meshes[m].points.size(); p++)
        {
            vector3 vertex = meshes[m].points[p].position;

            if (vertex.x < aabb[0]) aabb[0] = vertex.x;
            if (vertex.x > aabb[1]) aabb[1] = vertex.x;
            if (vertex.y < aabb[2]) aabb[2] = vertex.y;
            if (vertex.y > aabb[3]) aabb[3] = vertex.y;
            if (vertex.z < aabb[4]) aabb[4] = vertex.z;
            if (vertex.z > aabb[5]) aabb[5] = vertex.z;
        }
    }

    centre_of_bounding_sphere = (vector3(aabb[0], aabb[2], aabb[4]) + vector3(aabb[1], aabb[3], aabb[5])) / 2.0f;

    // Compute the radius of the bounding sphere for the model
    radius_of_bounding_sphere = 0.0f;

    // For each mesh
    for (m = 0; m < number_of_meshes; m++)
    {
        // For each point
        for (unsigned int p = 0; p < meshes[m].points.size(); p++)
        {
            vector3 vertex = meshes[m].points[p].position;

            float d = get_distance(vertex, centre_of_bounding_sphere);

            if (d > radius_of_bounding_sphere)
            {
                radius_of_bounding_sphere = d;
            }
        }
    }
}

ANIMATION_INSTANCE_HANDLE MilkshapeAscii::CreateAnimationInstance()
{
    ANIMATION_INSTANCE_HANDLE handle = new float[GetVertices().size() * 6];

    return handle;
}

void MilkshapeAscii::DeleteAnimationInstance(ANIMATION_INSTANCE_HANDLE instance)
{
    delete[] instance;
}

//! Returns the model geometry.
std::vector<vector3> MilkshapeAscii::GetVertices()
{
    std::vector<vector3> vertices;

    int number_of_meshes = meshes.size();

    // Go through each mesh
    for (int m = 0; m < number_of_meshes; m++)
    {
        for (int v = 0; v < meshes[m].vertex_count; v++)
        {
            int f = FloatsInFormat(meshes[m].vertex_format);

            vertices.push_back( vector3(meshes[m].vertex_array[v * f + 0],
                                        meshes[m].vertex_array[v * f + 1],
                                        meshes[m].vertex_array[v * f + 2]));
        }
    }

    return vertices;
}

//! Returns the indices for the model geometry.
std::vector<unsigned short> MilkshapeAscii::GetIndices()
{
    std::vector<unsigned short> indices;

    int number_of_meshes = meshes.size();

    // Go through each mesh
    for (int m = 0; m < number_of_meshes; m++)
    {
        short index_offset = -1;
        for (unsigned int v = 0; v < indices.size(); v++)
        {
            if (indices[v] > index_offset) index_offset = indices[v];
        }
        index_offset++;

        for (int i = 0; i < meshes[m].index_count; i++)
        {
            indices.push_back(meshes[m].index_array[i] + index_offset);
        }
    }

    return indices;
}

//! Updates an animated model
void MilkshapeAscii::Update(float time, ANIMATION_INSTANCE_HANDLE instance)
{
    if (instance == NULL)
    {
        std::cout << "Error: NULL passed to Update() function" << std::endl;
        exit(1);
    }


    int number_of_bones = skeleton.size();
    for (int b = 0; b < number_of_bones; b++)
    {
        int position_key_count = skeleton[b].position_keyframes.size();
        int rotation_key_count = skeleton[b].rotation_keyframes.size();
        if (position_key_count == 0 && rotation_key_count == 0)
        {
            skeleton[b].final = skeleton[b].absolute;
        }
        else
        {
            int i;
            std::vector<Keyframe> position_keyframes = skeleton[b].position_keyframes;
            std::vector<Keyframe> rotation_keyframes = skeleton[b].rotation_keyframes;
            vector3 pos, rot;

            // Find adjacent position keys
            int prev_position_keyframe = 0, next_position_keyframe = 0;
            for (i = 0; i < position_key_count; i++)
            {
                if (position_keyframes[i].time >= time)
                {
                    next_position_keyframe = i;
                    break;
                }
                prev_position_keyframe = i;
            }

            if (prev_position_keyframe != 0 && next_position_keyframe != 0)
            {
                float d = position_keyframes[next_position_keyframe].time - position_keyframes[prev_position_keyframe].time;
                float s = (time - position_keyframes[prev_position_keyframe].time) / d;
                pos = position_keyframes[prev_position_keyframe].parameter + (position_keyframes[next_position_keyframe].parameter - position_keyframes[prev_position_keyframe].parameter) * s;
            }
            else if (prev_position_keyframe == 0)
            {
                pos = position_keyframes[next_position_keyframe].parameter;
            }
            else if (next_position_keyframe == 0)
            {
                pos = position_keyframes[prev_position_keyframe].parameter;
            }

            // Find adjacent rotation keys
            matrix44 m;
            int prev_rotation_keyframe = 0, next_rotation_keyframe = 0;
            for (i = 0; i < rotation_key_count; i++)
            {
                if (rotation_keyframes[i].time >= time)
                {
                    next_rotation_keyframe = i;
                    break;
                }
                prev_rotation_keyframe = i;
            }
            if (prev_rotation_keyframe != 0 && next_rotation_keyframe != 0)
            {
                float d = rotation_keyframes[next_rotation_keyframe].time - rotation_keyframes[prev_rotation_keyframe].time;
                float s = (time - rotation_keyframes[prev_rotation_keyframe].time) / d;

                quaternion q1, q2, q;
                angle_quaternion (rotation_keyframes[prev_rotation_keyframe].parameter, q1);
                angle_quaternion (rotation_keyframes[next_rotation_keyframe].parameter, q2);
                quaternion_slerp (q1, q2, s, q);
                m = quaternion_matrix (q);
            }
            else if (prev_rotation_keyframe == 0)
            {
                rot[0] = RadToDeg(rotation_keyframes[next_rotation_keyframe].parameter[0]);
                rot[1] = RadToDeg(rotation_keyframes[next_rotation_keyframe].parameter[1]);
                rot[2] = RadToDeg(rotation_keyframes[next_rotation_keyframe].parameter[2]);
                m = angle_matrix(rot);
            }
            else if (next_rotation_keyframe == 0)
            {
                rot[0] = RadToDeg(rotation_keyframes[prev_rotation_keyframe].parameter[0]);
                rot[1] = RadToDeg(rotation_keyframes[prev_rotation_keyframe].parameter[1]);
                rot[2] = RadToDeg(rotation_keyframes[prev_rotation_keyframe].parameter[2]);
                m = angle_matrix(rot);
            }
            m[3][0] = pos[0];
            m[3][1] = pos[1];
            m[3][2] = pos[2];
            skeleton[b].relative_final = concat_transforms (skeleton[b].relative, m);
            int parent_bone = skeleton[b].parent;
            if (skeleton[b].parent == -1)
            {
                skeleton[b].final = skeleton[b].relative_final;
            }
            else
            {
                skeleton[b].final = concat_transforms(skeleton[parent_bone].final, skeleton[b].relative_final);
            }
        }
    }

    // Go through each mesh
    int vertex_number = 0;
    int number_of_meshes = meshes.size();
    for (int m = 0; m < number_of_meshes; m++)
    {
        int f = FloatsInFormat(meshes[m].vertex_format);

        for (int v = 0; v < meshes[m].vertex_count; v++)
        {
            int bone_index = meshes[m].points[meshes[m].point_indices[v]].bone_index;

            vector3 vertex, normal;
            vertex.x = meshes[m].vertex_array[v * f + 0];
            vertex.y = meshes[m].vertex_array[v * f + 1];
            vertex.z = meshes[m].vertex_array[v * f + 2];
            normal.x = meshes[m].vertex_array[v * f + 3];
            normal.y = meshes[m].vertex_array[v * f + 4];
            normal.z = meshes[m].vertex_array[v * f + 5];

            if (bone_index == -1)
            {
                instance[vertex_number * 6 + 0] = vertex.x;
                instance[vertex_number * 6 + 1] = vertex.y;
                instance[vertex_number * 6 + 2] = vertex.z;

                instance[vertex_number * 6 + 3] = normal.x;
                instance[vertex_number * 6 + 4] = normal.y;
                instance[vertex_number * 6 + 5] = normal.z;

            }
            else
            {
                vector3 v_tmp;
                // Transform vertices
                v_tmp = rotate(vertex, skeleton[bone_index].final);
                v_tmp[0] += skeleton[bone_index].final[3][0];
                v_tmp[1] += skeleton[bone_index].final[3][1];
                v_tmp[2] += skeleton[bone_index].final[3][2];

                instance[vertex_number * 6 + 0] = v_tmp[0];
                instance[vertex_number * 6 + 1] = v_tmp[1];
                instance[vertex_number * 6 + 2] = v_tmp[2];

                // Transform normals
                v_tmp = rotate(normal, skeleton[bone_index].final);

                instance[vertex_number * 6 + 3] = v_tmp[0];
                instance[vertex_number * 6 + 4] = v_tmp[1];
                instance[vertex_number * 6 + 5] = v_tmp[2];
            }

            vertex_number++;
        }
    }
}

//! Draws this model on screen
void MilkshapeAscii::Render(ANIMATION_INSTANCE_HANDLE instance, Light current_light, matrix44 object_to_world_transform)
{
    if (IGS == NULL)
    {
        throw std::string("Error: IGS == NULL in Render()");
    }

    IGS->SetCullMode(CULL_CW);

    int number_of_meshes = meshes.size();

    int vertex_number = 0;
    // Render each mesh
    for (int m = 0; m < number_of_meshes; m++)
    {
        // Set the appropriate material
        int mat_index = meshes[m].material_index;

        if (materials.size() != 0)
        {
            // Set the material
            IGS->SetMaterial(materials[mat_index].material);
        }
        else
        {
            MATERIAL m;

            m.diffuse = COLOUR(0.6f, 0.6f, 0.6f, 1.0f);
            m.specular= COLOUR(0.2f, 0.2f, 0.2f, 1.0f);
            m.emissive = COLOUR(0.0f, 0.0f, 0.0f, 1.0f);
            m.shiny = 0.0f;

            // Set the material
            IGS->SetMaterial(m);
        }

        if (skeleton.size() != 0)
        {
            if (instance == NULL)
            {
                throw std::string("Error: NULL instance handle passed to Render() and model has skeleton");
            }
            else // (instance != NULL)
            {
                int f = FloatsInFormat(meshes[m].vertex_format);

                for (int v = 0; v < meshes[m].vertex_count; v++)
                {
                    // Animated vertices
                    meshes[m].vertex_array_rendered[v * f + 0] = instance[vertex_number * 6 + 0];
                    meshes[m].vertex_array_rendered[v * f + 1] = instance[vertex_number * 6 + 1];
                    meshes[m].vertex_array_rendered[v * f + 2] = instance[vertex_number * 6 + 2];
                    // Animated normals
                    meshes[m].vertex_array_rendered[v * f + 3] = instance[vertex_number * 6 + 3];
                    meshes[m].vertex_array_rendered[v * f + 4] = instance[vertex_number * 6 + 4];
                    meshes[m].vertex_array_rendered[v * f + 5] = instance[vertex_number * 6 + 5];

                    vertex_number++;
                }
            }
        }

        if (materials.size() == 0)
        {
            IGS->SetMultiTextureMode(NORMAL);
            IGS->SetTexture(0, 0);
            IGS->SetTexture(1, 0);

            if (skeleton.size() == 0)
            {
                // Simplest case. Model has no bump map or animated skeleton.

                // Render the static mesh
                IGS->DrawIndexedTriangles(TRIANGLE_LIST, meshes[m].vertex_buffer, meshes[m].vertex_count,
                    meshes[m].index_buffer, meshes[m].index_count);

            }
            else
            {
                // Model has animated skeleton but no bumpmap (or bumpmapping disabled).

                // Render the mesh
                IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, meshes[m].vertex_format, meshes[m].vertex_array_rendered,
                    meshes[m].vertex_count, meshes[m].index_array, meshes[m].index_count);
            }
        }
        else if (materials[mat_index].dot3bump_handle == 0)
        {
            IGS->SetMultiTextureMode(NORMAL);
            IGS->SetTexture(0, materials[mat_index].texture_handle);
            IGS->SetTexture(1, 0);

            IGS->SetTextureWrappingMode(0, CLAMP_TEXTURE);
            if (materials[mat_index].wrap_u) IGS->SetTextureWrappingMode(0, WRAP_TEXTURE_S);
            if (materials[mat_index].wrap_v) IGS->SetTextureWrappingMode(0, WRAP_TEXTURE_T);

            if (materials[mat_index].alpha_texture_name == "chrome")
            {
                IGS->SetUVGenerationMode(0, UV_SPHERE_MAPPING);
            }

            if (skeleton.size() == 0)
            {
                // Simplest case. Model has no bump map or animated skeleton.

                // Render the static mesh
                IGS->DrawIndexedTriangles(TRIANGLE_LIST, meshes[m].vertex_buffer, meshes[m].vertex_count,
                    meshes[m].index_buffer, meshes[m].index_count);

            }
            else
            {
                // Model has animated skeleton but no bumpmap (or bumpmapping disabled).

                // Render the mesh
                IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, meshes[m].vertex_format, meshes[m].vertex_array_rendered,
                    meshes[m].vertex_count, meshes[m].index_array, meshes[m].index_count);
            }

            if (materials[mat_index].alpha_texture_name == "chrome")
            {
                IGS->SetUVGenerationMode(0, UV_DEFAULT);
            }
        }
/*
        else
        {

            MATERIAL mat;
            mat.diffuse = COLOUR(0.0f, 0.0f, 0.0f, 1.0f);
            mat.specular = COLOUR(0.0f, 0.0f, 0.0f, 1.0f);
            mat.emissive = COLOUR(0.0f, 0.0f, 0.0f, 1.0f);
            mat.shiny = 0.0f;

            // Set the material
            IGS->SetMaterial(materials[mat_index].material);


            IGS->SetMultiTextureMode(NORMAL);
            IGS->SetTexture(0, materials[mat_index].texture_handle);
            IGS->SetTexture(1, 0);

            // Render the mesh
            IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, meshes[m].vertex_format, meshes[m].vertex_array_rendered,
                meshes[m].vertex_count, meshes[m].index_array, meshes[m].index_count);

            // Blend in the Ambient lighting.
            IGS->SetBlendMode(ONE_ONE);

            IGS->SetMultiTextureMode(DOT3_BUMP);
            IGS->SetTexture(0, materials[mat_index].dot3bump_handle);
            IGS->SetTexture(1, materials[mat_index].texture_handle);

            // Model has a bump map. The vertex colours must be calculated from the light direction.

            // Get light direction in object space
            vector3 light_vec = light_vector.normalize();
            vector3 light_vec_reverse = reverse_rotate(light_vec, object_to_world_transform).normalize();

            int f = FloatsInFormat(meshes[m].vertex_format);
            int normal_offset = 3;
            int colour_offset = 6;

            float lv[3], lv_ts[3];
            lv[0] = light_vec_reverse[0];
            lv[1] = light_vec_reverse[1];
            lv[2] = light_vec_reverse[2];

            //convert to tangent space
            vecMat3x3Mult(lv, tangentMatrix, lv_ts);

            //scale and bias
            lv_ts[0] = lv_ts[0] * 0.5 + 0.5;
            lv_ts[1] = lv_ts[1] * 0.5 + 0.5;
            lv_ts[2] = lv_ts[2] * 0.5 + 0.5;

            // Compute L for each vertex
            for (int v = 0; v < meshes[m].vertex_count; v++)
            {
                vector3 normal;
                meshes[m].vertex_array_rendered[v * f + 6] = lv_ts[0];
                meshes[m].vertex_array_rendered[v * f + 7] = lv_ts[1];
                meshes[m].vertex_array_rendered[v * f + 8] = lv_ts[2];
            }

            // Render the mesh
            IGS->DrawIndexedTrianglesUP(TRIANGLE_LIST, meshes[m].vertex_format, meshes[m].vertex_array_rendered,
                meshes[m].vertex_count, meshes[m].index_array, meshes[m].index_count);

            IGS->SetBlendMode(BLEND_DISABLE);

        }
*/
    }
}

//! Determines whether or not the model is in the view frustum. time is only applicable to animated models.
bool MilkshapeAscii::IsWithinFrustum(matrix44 object_to_world_transform, float time = 0.0f)
{
    vector3 centre_transformed = transform(centre_of_bounding_sphere, object_to_world_transform);

    return SphereInFrustum(centre_transformed, radius_of_bounding_sphere);
}

vector3 project_ray(const vector3 &a, const vector3 &b, const vector3 &c,
    const vector3 &rOrigin, const vector3 &rVector)
{
    // Make the plane containing this triangle.
    vector3 pOrigin = a;
    vector3 v1 = b-a;
    vector3 v2 = c-a;

    // determine normal to plane containing polygon
    vector3 pNormal = CrossProduct(v1, v2).normalize();

    // find how far away the plane is from point rOrigin along the ray
    float distToPlaneIntersection = intersectRayPlane(rOrigin, rVector, pOrigin, pNormal);

    // find the nearest point on the plane to rOrigin
    return(rOrigin + rVector * distToPlaneIntersection);
}

//! Determines whether or not camera is shadowed by this object.
bool MilkshapeAscii::IsCameraWithinShadow(matrix44 object_to_world_transform, vector3 camera, Light current_light, float time)
{
    if (current_light.type == Light::DIRECTIONAL)
    {
        matrix44 identity = IdentityMatrix44();
        if (   object_to_world_transform[0][0] == identity[0][0]
            && object_to_world_transform[0][1] == identity[0][1]
            && object_to_world_transform[0][2] == identity[0][2]
            && object_to_world_transform[0][3] == identity[0][3]
            && object_to_world_transform[1][0] == identity[1][0]
            && object_to_world_transform[1][1] == identity[1][1]
            && object_to_world_transform[1][2] == identity[1][2]
            && object_to_world_transform[1][3] == identity[1][3]
            && object_to_world_transform[2][0] == identity[2][0]
            && object_to_world_transform[2][1] == identity[2][1]
            && object_to_world_transform[2][2] == identity[2][2]
            && object_to_world_transform[2][3] == identity[2][3])
        {
            if (current_light.direction.y < 0.0)
            {
                if (aabb[3] + object_to_world_transform[3][1] < camera.y)
                {
                    return false;
                }
            }
        }

        vector3 centre_transformed = transform(centre_of_bounding_sphere, object_to_world_transform);

        vector3 centre_projected = project_ray(current_light.clip1, current_light.clip2, current_light.clip3,
            centre_transformed, current_light.direction);

        // Determine the closest point to the camera on the line segment connecting centre_transformed and centre_projected
        vector3 closest = closestPointOnLine(centre_transformed, centre_projected, camera);

        return (get_distance(closest, camera) < radius_of_bounding_sphere);
    }
    else
    {
        return true;
    }
}

//! Determine if the shadow is within the view frustum.
bool MilkshapeAscii::IsShadowWithinFrustum(matrix44 object_to_world_transform, Light current_light, float time)
{
    if (current_light.type == Light::DIRECTIONAL)
    {
        vector3 centre_transformed = transform(centre_of_bounding_sphere, object_to_world_transform);

        vector3 centre_projected = project_ray(current_light.clip1, current_light.clip2, current_light.clip3,
            centre_transformed, current_light.direction);

        vector3 midpoint = (centre_transformed + centre_projected) / 2.0f;

        if (SphereInFrustum(midpoint, radius_of_bounding_sphere + get_distance(centre_transformed, centre_projected) / 2.0f))
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return true;
    }
}

//! Returns the radius and centre of the bounding sphere
void MilkshapeAscii::GetBoundingSphere(float& radius, vector3& centre)
{
    radius = radius_of_bounding_sphere;
    centre = centre_of_bounding_sphere;
}






