// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef MILKSHAPE_ASCII_MODEL_H
#define MILKSHAPE_ASCII_MODEL_H

/*! 
    \file milkshapeascii.h
    \brief Milkshape 3D ASCII file loader.
    File containing interface for loading a Milkshape3D ASCII file which 
    inherits from the abstract Model interface. 
*/

#include "model.h"
#include "transform.h"

//! Milkshape Ascii interface class
/*!
    Interface for Milkshape ASCII, inherits from the abstract base interface.
*/ 
class MilkshapeAscii : public IModel
{
public:
    //! Initialises the model.
    MilkshapeAscii();

    //! Deinitialises the model.
    virtual ~MilkshapeAscii();

    //! Loads a model from file.
    virtual void Load(std::string directory, std::string filename, IGraphicsSystem* IGS);

    //! Creates an animation instance for an animated model.
    virtual ANIMATION_INSTANCE_HANDLE CreateAnimationInstance();

    //! Deletes an animation instance.
    virtual void DeleteAnimationInstance(ANIMATION_INSTANCE_HANDLE instance);

    //! Updates an animated model
    virtual void Update(float time, ANIMATION_INSTANCE_HANDLE instance);

    //! Returns the model geometry.
    virtual std::vector<vector3> GetVertices();

    //! Returns the indices for the model geometry.
    virtual std::vector<unsigned short> GetIndices();

    //! Draws this model on screen
    virtual void Render(ANIMATION_INSTANCE_HANDLE instance, Light current_light, matrix44 object_to_world_transform);

    //! Determines whether or not the model is in the view frustum. time is only applicable to animated models.
    virtual bool IsWithinFrustum(matrix44 object_to_world_transform, float time);

    //! Determines whether or not camera is shadowed by this object.
    virtual bool IsCameraWithinShadow(matrix44 object_to_world_transform, vector3 camera, Light current_light, float time);

    //! Determine if the shadow is within the view frustum.
    virtual bool IsShadowWithinFrustum(matrix44 object_to_world_transform, Light current_light, float time);

    //! Returns the radius and centre of the bounding sphere
    virtual void GetBoundingSphere(float& radius, vector3& centre);

private:

    // Number of frames.
    int num_frames;

    // Frame number
    int frame;

    struct PointUV
    {
        int vflags;
        vector3 position;
        float u, v;
        int bone_index;
    };

    struct Face
    {
        int fflags;
        int v1;
        int v2;
        int v3;
        int n1;
        int n2;
        int n3;
        int smoothing_group;
    };

    struct Mesh
    {
        std::string mesh_name;          // Name of the mesh
        int mflags;
        int material_index;

        std::vector<PointUV> points;    // vertices in the mesh
        std::vector<vector3> normals;   // vertex normals
        std::vector<Face> faces;        // polygons in the mesh.


        int vertex_format;
        std::vector<int> point_indices;
        std::vector<int> normal_indices;


        // Number of vertices in the vertex buffer and vertex array & number of indices
        // in the index buffer and index array
        int vertex_count, index_count;

        // Index buffer and vertex buffer used if mesh is static
        VERTEX_BUFFER_HANDLE vertex_buffer;
        INDEX_BUFFER_HANDLE index_buffer;

        // Index array and vertex array used if buffer is dynamic
        unsigned short* index_array;
        float* vertex_array;
        float* vertex_array_rendered;
    };

    // Group of meshes in the model.
    std::vector<Mesh> meshes;

    struct MilkShapeMaterial
    {
        MATERIAL material;
        float transparency;
        std::string texture_name;
        std::string alpha_texture_name;

        bool wrap_u;
        bool wrap_v;

        TEXTURE_HANDLE texture_handle;  // Only valid if texture_name != ""
        TEXTURE_HANDLE dot3bump_handle; // Set to 0 if this is invalid
    };

    // Group of materials in the model.
    std::vector<MilkShapeMaterial> materials;

    // Animation keyframe information
    struct Keyframe
    {
        float time;   // in milliseconds
        vector3 parameter;
    };

    // Bone joint in the skeleton.
    struct Bone
    {
        std::string name;
        std::string parent_name;

        int flags;
        vector3 position;
        vector3 rotation;

        matrix44 relative, absolute, relative_final, final;

        std::vector<Keyframe> position_keyframes;
        std::vector<Keyframe> rotation_keyframes;

        int parent;
    };

    // The model's skeleton
    std::vector<Bone> skeleton;

    // Radius of the bounding sphere of the model
    float radius_of_bounding_sphere;
    vector3 centre_of_bounding_sphere;

    // Axis-aligned bounding box of the model.
    float aabb[6];
};

#endif // MILKSHAPE_ASCII_MODEL_H

