// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "SDL.h"
#include "SDL_thread.h"
#include <iostream>
#include "ode_thread.h"
#include "gettime.h"

SDL_mutex* ode_mutex;
SDL_Thread* ode_thread;

matrix44 ConvertToMatrix44(const dReal pos[3], const dReal R[12])
{
    matrix44 m;
    m.col[0].x = R[0];
    m.col[0].y = R[4];
    m.col[0].z = R[8];
    m.col[0].w = 0;
    m.col[1].x = R[1];
    m.col[1].y = R[5];
    m.col[1].z = R[9];
    m.col[1].w = 0;
    m.col[2].x = R[2];
    m.col[2].y = R[6];
    m.col[2].z = R[10];
    m.col[2].w = 0;
    m.col[3].x = pos[0];
    m.col[3].y = pos[1];
    m.col[3].z = pos[2];
    m.col[3].w = 1;

    return m;
}

// Given a body and a point in it's local coordinate system (centre of body is origin),
// return that point in the world's coordinate system ( (0,0,0) is origin ).
vector3 ConvertToWorldSpace(const vector3 &local_point, dBodyID body)
{
    dVector3 result;

    dBodyGetRelPointPos(body, local_point.x, local_point.y, local_point.z, result);

    return vector3(result[0], result[1], result[2]);
}

OdeThread::OdeThread()
{
    // Initialise ODE
    world = dWorldCreate();
    space = dHashSpaceCreate(0);
    contactgroup = dJointGroupCreate(0);
    dWorldSetGravity(world, 0, -0.5, 0);
    dWorldSetCFM(world, 1e-5);
    dWorldSetContactMaxCorrectingVel(world, 5.0);
    dWorldSetContactSurfaceLayer(world, 0.001);

    thread_exit = false;

    ode_mutex = SDL_CreateMutex();
}

OdeThread::~OdeThread()
{
    dJointGroupDestroy(contactgroup);
    dSpaceDestroy(space);
    dWorldDestroy(world);
}

// Starts the thread
void OdeThread::start_thread()
{
    ode_thread = SDL_CreateThread(ode_thread_function, this);
    if (ode_thread == NULL)
    {
        std::cout << "error creating thread." << std::endl;
    }
}

// Stops the thread
void OdeThread::stop_thread()
{
    ode_mutex_lock();

    thread_exit = true;

    ode_mutex_unlock();

    SDL_WaitThread(ode_thread, NULL);
}

// Returns the ODE world object
dWorldID OdeThread::getWorld()
{
    ode_mutex_lock();

    dWorldID temp = world;

    ode_mutex_unlock();

    return temp;
}

// Returns the ODE space object
dSpaceID OdeThread::getSpace()
{
    ode_mutex_lock();

    dSpaceID temp = space;

    ode_mutex_unlock();

    return temp;
}

void OdeThread::ode_mutex_lock()
{
#if defined(PHYSICS_IN_SEPARATE_THREAD)
    if (SDL_mutexP(ode_mutex) == -1)
    {
        std::cout << "Couldn't lock mutex." << std::endl;
    }
#endif // defined(PHYSICS_IN_SEPARATE_THREAD)
}

void OdeThread::ode_mutex_unlock()
{
#if defined(PHYSICS_IN_SEPARATE_THREAD)
    if (SDL_mutexV(ode_mutex) == -1)
    {
        std::cout << "Couldn't lock mutex." << std::endl;
    }
#endif // defined(PHYSICS_IN_SEPARATE_THREAD)
}

// this is called by dSpaceCollide when two objects in space are potentially colliding.
void nearCallback(OdeThread* ot, dGeomID o1, dGeomID o2)
{
    if (dGeomIsSpace(o1) || dGeomIsSpace(o2))
    {
        // colliding a space with something
        dSpaceCollide2(o1, o2, ot, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);

        // collide all geoms internal to the space(s)
        if (dGeomIsSpace (o1)) dSpaceCollide((dSpaceID)o1, ot, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);
        if (dGeomIsSpace (o2)) dSpaceCollide((dSpaceID)o2, ot, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);
    }
    else
    {

        dBodyID b1 = dGeomGetBody(o1);
        dBodyID b2 = dGeomGetBody(o2);

        // return without doing anything if both bodies are part of the static environment.
        if (b1 == 0 && b2 == 0) return;

        // return without doing anything if the two bodies are connected by a joint
        if (b1 && b2 && dAreConnectedExcluding(b1, b2, dJointTypeContact)) return;

        dContact contact[6];   // up to 6 contacts
        if (int numc = dCollide (o1, o2, 6, &contact[0].geom, sizeof(dContact)))
        {
            for (int i=0; i < numc; i++)
            {
                contact[i].surface.mode = dContactSlip1 | dContactSlip2 | dContactSoftERP | dContactSoftCFM | dContactApprox1;
                contact[i].surface.mu = dInfinity;
                contact[i].surface.slip1 = 0.1;
                contact[i].surface.slip2 = 0.1;
                contact[i].surface.soft_erp = 0.5;
                contact[i].surface.soft_cfm = 0.3;

                dJointID c = dJointCreateContact (ot->world, ot->contactgroup, contact+i);

                dJointAttach (c,b1,b2);
            }
        }
    }
}



void OdeThread::Update()
{
    const int updates_per_second = UPDATES_PER_SECOND;

    int update_time = 1000 / updates_per_second;
    if (update_time == 0) update_time = 1;

    static long start = GetTime();

    static long remaining_time = 0;

    remaining_time += GetTime() - start;
    start = GetTime();

    if (remaining_time > 250)
    {
        // More than a 250ms have elapsed.
        // This could be caused by the Operating System chugging away at some task or by the
        // user ALT+TABBING away, or by a v. slow framerate ( < 4 fps ). In any case we don't
        // update the physics engine.
        remaining_time = 0;
    }

    while (remaining_time > update_time)
    {
        dSpaceCollide(space, this, (void(*)(void *, struct dxGeom *, struct dxGeom *))&nearCallback);

        float step_size = 7.0f / float(updates_per_second);

        dWorldQuickStep(world, step_size);

        // Remove all contact joints
        dJointGroupEmpty (contactgroup);

        remaining_time -= update_time;
    }
}

int ode_thread_function(void* void_ptr)
{
#if defined(PHYSICS_IN_SEPARATE_THREAD)
    OdeThread* ot = (OdeThread*)void_ptr;

    ot->ode_mutex_lock();

    while (ot->thread_exit == false)
    {
        ot->Update();

        ot->ode_mutex_unlock();

        // Sleep for 5 milliseconds.
        SleepMS(5);

        ot->ode_mutex_lock();
    }

    ot->ode_mutex_unlock();
#endif // defined(PHYSICS_IN_SEPARATE_THREAD)
    return 0;
}

