// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#ifndef ODE_THREAD_H
#define ODE_THREAD_H

#include <ode/ode.h>
#include "mtxlib.h"

/*!
    \file ode_thread.h
    \brief File containing class that deals with ODE.
*/

//! Determines whether to update the physics in a thread of its own.
#define PHYSICS_IN_SEPARATE_THREAD
//! Number of physics updates per second.
#define UPDATES_PER_SECOND 80

// Converts an ODE position and 3x3 rotation matrix to matrix44 format
matrix44 ConvertToMatrix44(const dReal pos[3], const dReal R[12]);

// Given a body and a point in it's local coordinate system (centre of body is origin),
// return that point in the world's coordinate system ( (0,0,0) is origin ).
vector3 ConvertToWorldSpace(const vector3 &local_point, dBodyID body);

//! Class that deals with the thread for the physics code.
class OdeThread
{
public:
    //! Returns a reference to the one and only instance of this class.
    static OdeThread& Instance()
    {
        static OdeThread Obj;
        return Obj;
    }

    //! Destructor.
    /*!
        Destroys world and space
    */
    ~OdeThread();

    //! Starts the thread
    /*!
        Starts the ODE thread.
    */
    void start_thread();

    //! Stops the thread
    /*!
        Stops the ODE thread.
    */
    void stop_thread();

    //! Returns the ODE world object
    /*!
        When calling this function ensure that the calling thread doesn't have the mutex locked.
        \return the ODE dWorldID for the main world
    */
    dWorldID getWorld();

    //! Returns the ODE space object
    /*!
        When calling this function ensure that the calling thread doesn't have the mutex locked.
        \return the ODE dSpaceID for the main space
    */
    dSpaceID getSpace();

    //! Locks the thread mutex.
    /*!
        This mutex is to prevent two threads accessing ODE at the same time.
        Call this before calling ODE functions and then call ode_mutex_unlock after using the ODE functions
    */
    static void ode_mutex_lock();

    //! Unocks the thread mutex.
    /*!
        This mutex is to prevent two threads accessing ODE at the same time.
        Call this after calling ode_mutex_lock followed by some ODE functions.
    */
    static void ode_mutex_unlock();

    // Updates the physics timestep(s).
    /*!
        If PHYSICS_IN_SEPARATE_THREAD is not defined this function should be called once per frame.
    */
    void Update();

private:

    // Constructor is private in this singleton
    OdeThread();

    // ODE specific stuff
    dJointGroupID contactgroup;
    dWorldID world;
    dSpaceID space;

    // This is set to true to tell the thread to stop.
    bool thread_exit;

    //! ODE Collision callback function
    /*!
        Callback called by ODE when two objects are potentially colliding.
        \param ot pointer to the OdeThread class.
        \param o1 dGeomID of one geometry object
        \param o2 dGeomID of one geometry object
    */
    friend void nearCallback(OdeThread* ot, dGeomID o1, dGeomID o2);

    //! Thread function
    /*!
        Loops repeatedly calling Update() whenever the physics need to be updated. Sleeps for 5ms if
        physics don't need to be updated.
    */
    friend int ode_thread_function(void*);
};

#endif // ODE_THREAD_H

