// Slickworm
// Copyright (c) 2002-2004, Michel O'Brien.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  1. Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//  2. Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//  3. Neither the names of the copyright holders nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include "particle.h"
#include "modelfactory.h"
#include "object_renderer.h"
#include "gettime.h"

Particle::Particle(IGraphicsSystem* IGS, dWorldID world, dSpaceID space, vector3 position, vector3 velocity,
    std::string model_directory, std::string model_filename, IModel* model, object_type type, vector3 size)
{
    OdeThread::ode_mutex_lock();

    this->world = world;
    this->space = space;
    this->IGS = IGS;

    // Create the bodies for the sphere
    #define DENSITY (1.0)   // Density of all objects
    dMass m;

    body = dBodyCreate(world);

    if (type == SPHERE)
    {
        float radius = size.x;

        dMassSetSphere(&m, DENSITY, radius);
        dMassAdjust (&m, 1.0);

        dBodySetMass(body, &m);
        geom = dCreateSphere(space, radius);
        dGeomSetBody(geom, body);
    }
    else if (type == BOX)
    {
        float x_size = size.x;
        float y_size = size.y;
        float z_size = size.z;

        dMassSetBox(&m, DENSITY, x_size, y_size, z_size);
        dMassAdjust(&m, 1.0);

        dBodySetMass(body, &m);
        geom = dCreateBox(space, x_size, y_size, z_size);
        dGeomSetBody(geom, body);
    }
    else // (type == CAPSULE)
    {
        float length = size.x;
        float radius = size.y;

        dMassSetCappedCylinder(&m, DENSITY, 1, radius, length);
        dMassAdjust(&m, 1.0);

        dBodySetMass(body, &m);
        geom = dCreateCCylinder (space, radius, length);
        dGeomSetBody(geom, body);
    }

    dBodySetPosition(body, position.x, position.y, position.z);
    dBodySetLinearVel(body, velocity.x, velocity.y, velocity.z);

    if (model == NULL)
    {
        this->model = ModelFactory::CreateModel(ModelFactory::MILKSHAPE_ASCII);
        this->model->Load(model_directory, model_filename, IGS);

        this->delete_model = true;
    }
    else // model != NULL)
    {
        this->model = model;

        this->delete_model = false;
    }

    // Store the creation time of this particle.
    spawn_time = GetTime();

    OdeThread::ode_mutex_unlock();
}

Particle::~Particle()
{
    OdeThread::ode_mutex_lock();

    if (this->delete_model)
    {
        delete model;
    }
    dGeomDestroy(geom);
    dBodyDestroy(body);

    OdeThread::ode_mutex_unlock();
}

void Particle::Render()
{
    matrix44 translation_body = ConvertToMatrix44(dBodyGetPosition(body), dBodyGetRotation(body));

    ObjectRenderer::Instance().AddModel(model, translation_body);
}

vector3 Particle::GetPosition()
{
    vector3 result;

    OdeThread::ode_mutex_lock();
    result = ConvertToWorldSpace(vector3(0.0, 0.0, 0.0), body);
    OdeThread::ode_mutex_unlock();

    return result;
}

